import logging
from os import path
import tornado.web
from tornado.escape import url_escape, url_unescape

from temboardui.web import (
    Blueprint,
    HTTPError,
    Redirect,
    TemplateRenderer,
)


PLUGIN_NAME = 'pgconf'
logger = logging.getLogger(__name__)
blueprint = Blueprint()
blueprint.generic_proxy("/pgconf/configuration", methods=["POST"])
plugin_path = path.dirname(path.realpath(__file__))
render_template = TemplateRenderer(plugin_path + "/templates")


def configuration(config):
    return {}


def get_routes(config):
    routes = blueprint.rules + [
        (
            r"/js/pgconf/(.*)",
            tornado.web.StaticFileHandler,
            {'path': plugin_path + "/static/js"}
        ),
        (
            r"/css/pgconf/(.*)",
            tornado.web.StaticFileHandler,
            {'path': plugin_path + "/static/css"}
        ),
    ]
    return routes


@blueprint.instance_route("/pgconf/configuration(?:/category/(.+))?",
                          methods=["GET", "POST"])
def configuration_handler(request, category=None):
    request.instance.check_active_plugin(PLUGIN_NAME)
    profile = request.instance.get_profile()
    agent_username = profile['username']
    template_vars = {}
    # Deduplicate HTTP prefix of plugin on agent.
    prefix = "/pgconf/configuration"
    query_filter = request.handler.get_argument('filter', None, strip=True)

    status = request.instance.get(prefix + "/status")
    categories = request.instance.get(prefix + "/categories")

    if category:
        category = url_unescape(category)
    else:
        category = categories['categories'][0]
    logger.debug("category=%s", category)

    if query_filter:
        query = {'filter': query_filter}
        configuration_url = prefix
    else:
        query = {}
        configuration_url = prefix + "/category/" + url_escape(category)
    configuration = request.instance.get(configuration_url, query=query)

    if "POST" == request.method:
        settings = {'settings': [
            {'name': name, 'setting': value[0]}
            for name, value in request.arguments.iteritems()
            # 'filter' is not a setting, just ignore it.
            if name != 'filter'
        ]}
        try:
            request.instance.post(prefix, body=settings)
            # Redirect to GET page, same URI.
            return Redirect(request.uri)
        except HTTPError as e:
            # Rerender HTML page with errors.
            template_vars['error_code'] = e
            template_vars['error_message'] = e.log_message

    return render_template(
        'configuration.html',
        nav=True,
        role=request.current_user,
        instance=request.instance,
        agent_username=agent_username,
        plugin=PLUGIN_NAME,
        xsession=request.instance.xsession,
        current_cat=category,
        configuration_categories=categories,
        configuration_status=status,
        data=configuration,
        query_filter=query_filter,
        **template_vars
    )



from temboardui.application import (
    add_instance,
    add_instance_in_group,
    add_instance_plugin,
    check_agent_address,
    check_agent_port,
    delete_instance,
    delete_instance_from_group,
    get_group_list,
    get_groups_by_instance,
    get_instance_list,
    purge_instance_plugins,
    update_instance,
)
from temboardui.temboardclient import temboard_discover
from temboardui.web import (
    HTTPError,
    InstanceHelper,
    admin_required,
    app,
    render_template,
)


def add_instance_in_groups(db_session, instance, groups):
    for group_name in groups or []:
        add_instance_in_group(
            db_session, instance.agent_address, instance.agent_port,
            group_name)


def create_instance_helper(webapp, db_session, data):
    validate_instance_data(data)
    groups = data.pop('groups')
    plugins = data.pop('plugins') or []
    instance = add_instance(db_session, **data)
    add_instance_in_groups(db_session, instance, groups)
    enable_instance_plugins(
        db_session, instance, plugins, webapp.loaded_plugins,
    )


def enable_instance_plugins(db_session, instance, plugins, loaded_plugins):
    for plugin_name in plugins or []:
        # 'administration' plugin case: the plugin is not currently
        # implemented on UI side
        if plugin_name == 'administration':
            continue
        if plugin_name not in loaded_plugins:
            raise HTTPError(404, "Unknown plugin %s." % plugin_name)

        add_instance_plugin(
            db_session, instance.agent_address,
            instance.agent_port, plugin_name)


def validate_instance_data(data):
    # Submited attributes checking.
    if not data.get('new_agent_address'):
        raise HTTPError(400, "Agent address is missing.")
    check_agent_address(data['new_agent_address'])
    if 'new_agent_port' not in data or data['new_agent_port'] == '':
        raise HTTPError(400, "Agent port is missing.")
    check_agent_port(data['new_agent_port'])
    if 'agent_key' not in data:
        raise HTTPError(400, "Agent key field is missing.")
    if 'groups' not in data:
        raise HTTPError(400, "Groups field is missing.")
    if data['groups'] is not None and type(data['groups']) != list:
        raise HTTPError(400, "Invalid group list.")


@app.route(r"/json/settings/instance", methods=['POST'])
@admin_required
def create_instance_handler(request):
    create_instance_helper(
        request.handler.application, request.db_session, request.json,
    )
    return {"message": "OK"}


@app.route(
    r"/json/settings/instance" + InstanceHelper.INSTANCE_PARAMS,
    methods=['GET', 'POST'], with_instance=True)
@admin_required
def json_instance(request):
    instance = request.instance
    if 'GET' == request.method:
        groups = get_group_list(request.db_session, 'instance')
        return {
            'agent_address': instance.agent_address,
            'agent_port': instance.agent_port,
            'agent_key': instance.agent_key,
            'hostname': instance.hostname,
            'cpu': instance.cpu,
            'memory_size': instance.memory_size,
            'pg_port': instance.pg_port,
            'pg_version': instance.pg_version,
            'pg_data': instance.pg_data,
            'in_groups': [g.group_name for g in instance.groups],
            'enabled_plugins': [p.plugin_name for p in instance.plugins],
            'groups': [{
                'name': group.group_name,
                'description': group.group_description
            } for group in groups],
            'loaded_plugins': request.handler.application.loaded_plugins,
            'notify': instance.notify,
            'comment': instance.comment,
        }
    else:  # POST (update)
        data = request.json
        validate_instance_data(data)
        groups = data.pop('groups')
        plugins = data.pop('plugins') or []

        # First step is to remove the instance from the groups it belongs to.
        instance_groups = get_groups_by_instance(
            request.db_session, instance.agent_address,
            instance.agent_port)
        for instance_group in instance_groups:
            delete_instance_from_group(
                request.db_session, instance.agent_address,
                instance.agent_port, instance_group.group_name)
        # Remove plugins
        purge_instance_plugins(
            request.db_session, instance.agent_address,
            instance.agent_port)

        instance = update_instance(
            request.db_session,
            instance.agent_address,
            instance.agent_port,
            **data)
        add_instance_in_groups(request.db_session, instance, groups)
        enable_instance_plugins(
            request.db_session, instance, plugins,
            request.handler.application.loaded_plugins,
        )
        return {"message": "OK"}


@app.route(r"/json/settings/delete/instance$", methods=['POST'])
@admin_required
def json_delete_instance(request):
    data = request.json
    if not data.get('agent_address'):
        raise HTTPError(400, "Agent address field is missing.")
    if not data.get('agent_port'):
        raise HTTPError(400, "Agent port field is missing.")
    delete_instance(request.db_session, **data)
    return {'delete': True}


@app.route(
    r"/json/discover/instance" + InstanceHelper.INSTANCE_PARAMS)
@admin_required
def discover(request, address, port):
    return temboard_discover(
        request.config.temboard.ssl_ca_cert_file, address, port)


@app.route(r"/settings/instances")
@admin_required
def instances(request):
    return render_template(
        'settings/instance.html',
        nav=True, role=request.current_user,
        instance_list=get_instance_list(request.db_session)
    )


@app.route(r"/json/register/instance", methods=['POST'])
@admin_required
def register(request):
    data = request.json
    agent_address = data.pop('agent_address', None)
    if not agent_address:
        # Try to find agent's IP
        x_real_ip = request.headers.get("X-Real-IP")
        agent_address = x_real_ip or request.remote_ip

    data['new_agent_address'] = agent_address
    data['new_agent_port'] = data.pop('agent_port', None)
    create_instance_helper(
        request.handler.application,
        request.db_session,
        data,
    )
    return {"message": "OK"}



# -*- coding: utf-8 -*-
from datetime import datetime, timedelta
import json
import hashlib
import logging
import os
from textwrap import dedent
from urllib import quote
import urllib2

import tornado.web
import tornado.escape

from sqlalchemy.orm import sessionmaker, scoped_session
from sqlalchemy.exc import ProgrammingError, IntegrityError, DataError
from sqlalchemy.sql import text

from psycopg2.extensions import AsIs

from temboardui.toolkit import taskmanager
from temboardui.application import (
    get_roles_by_instance,
    send_mail,
    send_sms,
)
from temboardui.temboardclient import temboard_request
from temboardui.model.orm import Instances
from temboardui.model import worker_engine

from .model.orm import (
    Check,
    CollectorStatus,
    Host,
    Instance,
)
from .model.db import insert_availability
from .alerting import (
    check_specs,
)
from .handlers import blueprint
from .tools import (
    check_preprocessed_data,
    get_host_id,
    get_instance_checks,
    get_instance_id,
    insert_metrics,
    merge_agent_info,
    populate_host_checks,
    preprocess_data,
    update_collector_status,
)

logger = logging.getLogger(__name__)
PLUGIN_NAME = 'monitoring'
workers = taskmanager.WorkerSet()


def configuration(config):
    return {}


def get_routes(config):
    plugin_path = os.path.dirname(os.path.realpath(__file__))
    __import__(__name__ + '.handlers.alerting')
    __import__(__name__ + '.handlers.monitoring')
    routes = blueprint.rules + [
        (r"/js/monitoring/(.*)",
         tornado.web.StaticFileHandler, {'path': plugin_path + "/static/js"}),
    ]
    return routes


@workers.register(pool_size=1)
def aggregate_data_worker(app):
    # Worker in charge of aggregate data
    try:
        engine = worker_engine(app.config.repository)
        with engine.begin() as conn:
            conn.execute("SET search_path TO monitoring")
            logger.debug("Running SQL function monitoring.aggregate_data()")
            res = conn.execute("SELECT * FROM aggregate_data()")
            for row in res.fetchall():
                logger.debug("table=%s insert=%s"
                             % (row['tblname'], row['nb_rows']))
            return
    except Exception as e:
        logger.error('Could not aggregate montitoring data')
        logger.exception(e)
        raise(e)


@workers.register(pool_size=1)
def history_tables_worker(app):
    # Worker in charge of history tables
    try:
        engine = worker_engine(app.config.repository)
        with engine.connect() as conn:
            conn.execute("SET search_path TO monitoring")
            logger.debug("Running SQL function monitoring.history_tables()")
            res = conn.execute("SELECT * FROM history_tables()")
            for row in res.fetchall():
                logger.debug("table=%s insert=%s"
                             % (row['tblname'], row['nb_rows']))
            conn.execute("COMMIT")
            return
    except Exception as e:
        logger.error('Could not history montitoring tables')
        logger.exception(e)
        try:
            conn.execute("ROLLBACK")
        except Exception:
            pass
        raise(e)


@workers.register(pool_size=10)
def check_data_worker(app, host_id, instance_id, data):
    # Worker in charge of checking preprocessed monitoring values
    engine = worker_engine(app.config.repository)
    session_factory = sessionmaker(bind=engine)
    Session = scoped_session(session_factory)
    worker_session = Session()

    check_preprocessed_data(
        worker_session,
        host_id,
        instance_id,
        data,
        app.config.temboard.home
    )

    worker_session.commit()
    worker_session.close()


@workers.register(pool_size=1)
def purge_data_worker(app):
    """Background worker in charge of purging monitoring data. Purge policy
    is based on purge_after parameter from monitoring section. purger_after
    defines the number of day of data to keep, from now. Default value means
    there is no purge policy.
    """

    logger.setLevel(app.config.logging.level)
    logger.info("Starting monitoring data purge worker.")

    if not app.config.monitoring.purge_after:
        logger.info("No purge policy, end.")
        return

    engine = worker_engine(app.config.repository)

    with engine.connect() as conn:
        # Get tablename list to purge from metric_tables_config()
        res = conn.execute(
            dedent("""
                SELECT
                    tablename
                FROM (
                    SELECT
                        tablename_prefix||'_'||suffix AS tablename
                    FROM
                        json_object_keys(monitoring.metric_tables_config()) AS tablename_prefix,
                        UNNEST(ARRAY['30m_current', '6h_current', 'current', 'history']) AS suffix
                ) AS q
                WHERE EXISTS (
                    SELECT 1
                    FROM
                        pg_catalog.pg_tables AS pgt
                    WHERE
                        pgt.tablename=q.tablename
                        AND pgt.schemaname = 'monitoring'
                )
                ORDER BY tablename;
            """)  # noqa
        )
        tablenames = [r['tablename'] for r in res.fetchall()]
        tablenames.extend(['state_changes', 'check_changes'])

        purge_query_base = "DELETE FROM :tablename WHERE "

        for tablename in tablenames:

            # With history tables, we have to deal with tstzrange
            if tablename.endswith("_history"):
                query = purge_query_base + \
                        "NOT (history_range && tstzrange(NOW() " + \
                        "- ':nday days'::INTERVAL, NOW()))"
            else:
                query = purge_query_base + \
                        "datetime < (NOW() - ':nday days'::INTERVAL)"

            logger.debug("Purging table %s", tablename)
            t = conn.begin()
            try:
                res_delete = conn.execute(
                    text(query),
                    tablename=AsIs("monitoring.%s" % tablename),
                    nday=app.config.monitoring.purge_after,
                )
                t.commit()
            except (ProgrammingError, IntegrityError) as e:
                logger.exception(e)
                logger.error("Could not delete data from table %s", tablename)
                t.rollback()
                continue

            if res_delete.rowcount > 0:
                logger.info("Table %s purged, %s rows deleted",
                            tablename, res_delete.rowcount)

    logger.info("End of monitoring data purge worker.")


@workers.register(pool_size=1)
def notify_state_change(app, check_id, key, value, state, prev_state):
    # check if at least one notifications transport is configured
    # if it's not the case pass
    notifications_conf = app.config.notifications
    smtp_host = notifications_conf.smtp_host
    smtp_port = notifications_conf.smtp_port
    smtp_tls = notifications_conf.smtp_tls
    smtp_login = notifications_conf.smtp_login
    smtp_password = notifications_conf.smtp_password
    smtp_from_addr = notifications_conf.smtp_from_addr

    if not smtp_host and \
       not notifications_conf.get('twilio_account_sid', None):
        logger.info("No SMTP nor SMS service configured, "
                    "notification not sent")
        return

    # Worker in charge of sending notifications
    engine = worker_engine(app.config.repository)
    session_factory = sessionmaker(bind=engine)
    Session = scoped_session(session_factory)
    worker_session = Session()

    check = worker_session.query(Check).filter(
        Check.check_id == check_id,
    ).join(Instance).join(Host).one()

    port = check.instance.port
    hostname = check.instance.host.hostname
    instance = worker_session.query(Instances).filter(
        Instances.pg_port == port,
        Instances.hostname == hostname,
    ).one()

    # don't notify if notifications are disabled for this instance
    if not instance.notify:
        return

    specs = check_specs
    spec = specs.get(check.name)

    message = ''
    if state != 'OK':
        message = spec.get('message').format(
            key=key,
            check=check.name,
            value=value,
            threshold=getattr(check, state.lower()),
        )

    description = spec.get('description')
    subject = '[temBoard] {state} {hostname} - {description}' \
        .format(hostname=hostname, state=state, description=description)
    link = 'https://%s:%d/server/%s/%d/alerting/%s' % (
        app.config.temboard.address,
        app.config.temboard.port,
        instance.agent_address,
        instance.agent_port,
        check.name)

    direction = '➚' if prev_state == 'OK' or state == 'CRITICAL' else '➘'

    body = '''
    Instance: {hostname}:{port}
    Description: {description}
    Status: {direction} {state} (prev. {prev_state})
    {message}
    {link}
    '''.format(
        hostname=hostname,
        port=instance.agent_port,
        description=description,
        direction=direction,
        state=state,
        prev_state=prev_state,
        message=message,
        link=link,
    )

    roles = get_roles_by_instance(worker_session,
                                  instance.agent_address,
                                  instance.agent_port)

    emails = [role.role_email for role in roles if role.role_email]
    if len(emails):
        send_mail(smtp_host, smtp_port, subject, body, emails, smtp_tls,
                  smtp_login, smtp_password,  smtp_from_addr)

    phones = [role.role_phone for role in roles if role.role_phone]
    if len(phones):
        send_sms(app.config.notifications, body, phones)


@workers.register(pool_size=1)
def schedule_collector(app):
    """Worker function in charge of scheduling collector (pull mode)."""

    logger.setLevel(app.config.logging.level)
    logger.info("Starting collector scheduler worker.")

    engine = worker_engine(app.config.repository)
    with engine.connect() as conn:
        # Get the list of agents
        res = conn.execute(
            "SELECT agent_address, agent_port, agent_key "
            "FROM application.instances ORDER BY 1, 2"
        )

        for agent_address, agent_port, agent_key in res.fetchall():
            # For each registered agent, let's start a new data collector

            # Build a unique Task id based on agent address and port
            task_id = hashlib.md5(
                "%s:%s" % (agent_address, agent_port)
            ).hexdigest()[:8]

            taskmanager.schedule_task(
                'collector',
                listener_addr=os.path.join(
                    app.config.temboard.home, '.tm.socket'
                ),
                id=task_id,
                options=dict(
                    address=agent_address,
                    port=agent_port,
                    key=agent_key,
                ),
                expire=0,
            )

    logger.info("End of collector scheduler worker.")


@workers.register(pool_size=20)
def collector(app, address, port, key):
    logger.setLevel(app.config.logging.level)
    logger.info("Starting collector worker for %s:%s", address, port)

    discover_data = dict()

    # First, we need to call discover API because we want to know the hostname
    url = 'https://%s:%s/discover' % (address, port)
    logger.debug("Calling %s" % url)

    try:
        response = temboard_request(
            app.config.temboard.ssl_ca_cert_file,
            'GET',
            url,
            headers={"Content-type": "application/json"},
        )
    except urllib2.HTTPError:
        logger.error("Could not get response from %s" % url)
        logger.error("Agent or host are down.")
        return

    discover_data = json.loads(response)
    logger.debug(discover_data)

    # Start new ORM DB session
    engine = worker_engine(app.config.repository)
    session_factory = sessionmaker(bind=engine)
    Session = scoped_session(session_factory)
    worker_session = Session()

    host_id = instance_id = None
    # Agent monitoring API endpoint
    history_url = 'https://%s:%s/monitoring/history?key=%s&limit=100' % (
        address, port, quote(key)
    )

    try:
        # Trying to find host_id, instance_id and the datetime of the latest
        # inserted record.

        # Find host_id and instance_id by hostname and PG port
        host_id = get_host_id(worker_session, discover_data['hostname'])
        instance_id = get_instance_id(
            worker_session,
            host_id,
            discover_data['pg_port']
        )
    except Exception:
        # This case happens on the very first pull when no data have been
        # previously added.
        logger.warning(
            "Could not find host or instance records in monitoring inventory "
            "tables."
        )
    else:
        # Get last inserted data timestamp from collector status
        collector_status = worker_session.query(CollectorStatus).filter(
            CollectorStatus.instance_id == instance_id
        ).first()

        if collector_status and collector_status.last_insert:
            start = (
                collector_status.last_insert + timedelta(seconds=1)
            ).strftime("%Y-%m-%dT%H:%M:%SZ")
            history_url += "&start=%s" % start

    # Finally, let's call /monitoring/history agent API for getting metrics
    # history.
    try:
        logger.debug("Calling %s" % history_url)
        response = temboard_request(
            app.config.temboard.ssl_ca_cert_file,
            'GET',
            history_url,
            headers={"Content-type": "application/json"},
        )
    except urllib2.HTTPError as e:
        if e.code == 404:
            logger.error("Agent version does not support pull mode. "
                         "Please consider upgrading to version 5 at least.")
        else:
            logger.exception(str(e))
            # Update collector status only if instance_id is known
            if instance_id:
                update_collector_status(
                    worker_session,
                    instance_id,
                    u'FAIL',
                    last_pull=datetime.utcnow(),
                )
                worker_session.commit()
        return

    for row in json.loads(response):
        hostinfo = row['hostinfo']
        data = row['data']
        instance = row['instances'][0]
        port = instance['port']
        if 'max_connections' in instance:
            row['data']['max_connections'] = instance['max_connections']

        try:
            # Try to insert collected data

            # Update the inventory
            host = merge_agent_info(worker_session, hostinfo, instance)
            instance_id = get_instance_id(worker_session, host.host_id, port)
            # Insert instance availability informations
            insert_availability(
                worker_session,
                row['datetime'],
                instance_id,
                row['instances'][0]['available']
            )
            # Insert collected metrics
            insert_metrics(worker_session, host.host_id, instance_id, data)
            worker_session.commit()

        except DataError as e:
            # Wrong data type or corrupted data could lead to DataError. In
            # this case we should consider this row as unvalid and move to the
            # next one.
            try:
                last_insert = datetime.strptime(
                    row['datetime'], "%Y-%m-%d %H:%M:%S +0000"
                )
            except ValueError:
                # If row datetime could not be parsed, we should fallback to
                # current datetime. This will result to ignore potential valid
                # rows between row datetime and now, but this is better than
                # letting the collector stucked for ever on an invalid row.
                last_insert = datetime.utcnow()

            logger.exception(str(e))
            worker_session.rollback()
            # Set collector status to FAIL
            if instance_id:
                update_collector_status(
                    worker_session,
                    instance_id,
                    u'FAIL',
                    last_pull=datetime.utcnow(),
                    last_insert=last_insert,
                )
                worker_session.commit()
            # Continue with the next row
            continue

        # Update collector status
        update_collector_status(
            worker_session,
            instance_id,
            u'OK',
            last_pull=datetime.utcnow(),
            # This is the datetime format used by the agent
            last_insert=datetime.strptime(
                row['datetime'], "%Y-%m-%d %H:%M:%S +0000"
            ),
        )
        worker_session.commit()
        # ALERTING PART
        populate_host_checks(
            worker_session,
            host.host_id,
            instance_id,
            dict(n_cpu=hostinfo['cpu_count']),
        )
        worker_session.commit()

        # Apply alerting checks against preprocessed data
        check_preprocessed_data(
            worker_session,
            host.host_id,
            instance_id,
            preprocess_data(
                row['data'],
                get_instance_checks(worker_session, instance_id),
                row['datetime']
            ),
            app.config.temboard.home,
        )

        logger.debug("Row with datetime=%s inserted", row['datetime'])
        worker_session.commit()

    logger.info("End of collector worker.")


@taskmanager.bootstrap()
def monitoring_bootstrap(context):
    yield taskmanager.Task(
            worker_name='aggregate_data_worker',
            id='aggregate_data',
            redo_interval=30 * 60,  # Repeat each 30m,
            options={},
    )
    yield taskmanager.Task(
            worker_name='history_tables_worker',
            id='history_tables',
            redo_interval=3 * 60 * 60,  # Repeat each 3h
            options={},
    )
    yield taskmanager.Task(
            worker_name='purge_data_worker',
            id='purge_data',
            redo_interval=24 * 60 * 60,  # Repeat each 24h
            options={},
    )
    yield taskmanager.Task(
            worker_name='schedule_collector',
            id='schedule_collector',
            redo_interval=60,  # Repeat each 60s
            options={},
    )



from dateutil import parser as parse_datetime
import logging
import os

from sqlalchemy.orm.exc import NoResultFound

from temboardui.web import HTTPError
from temboardui.toolkit import taskmanager

from .model.orm import (
    Check,
    CheckState,
    CollectorStatus,
    Host,
    Instance,
)
from .model import db
from .alerting import (
    bootstrap_checks,
    check_specs,
)


logger = logging.getLogger(__name__)


def merge_agent_info(session, host_info, instance_info):
    """Update the host, instance and database information with the
    data received from the agent."""

    try:
        # Try to get host_id, based on hostname
        host_info['host_id'] = get_host_id(session, host_info['hostname'])
    except Exception:
        # host not found
        pass

    host = Host.from_dict(host_info)

    # Insert or update host information
    session.merge(host)
    session.flush()
    session.commit()

    # Get host_id in any case
    host_id = get_host_id(session, host_info['hostname'])

    # Only process instances marked as available, since only those
    # have complete information
    if instance_info['available']:
        try:
            # Try to get instance_id
            instance_info['instance_id'] = get_instance_id(
                session, host_id, instance_info['port']
            )
        except Exception:
            # instance not found
            pass
        instance_info['host_id'] = host_id

        inst = Instance.from_dict(instance_info)
        # Insert or update instance information
        session.merge(inst)
        session.flush()
        session.commit()
    return host


def get_host_id(session, hostname):
    """Get host_id by the hostname.
    """
    host_id = db.get_host_id(session, hostname)
    if not host_id:
        raise Exception(
            "Could not find registered host with hostname=%s" % hostname
        )
    return host_id


def get_instance_id(session, host_id, port):
    """Get instance_id by host_id and port.
    """
    instance_id = db.get_instance_id(session, host_id, port)
    if not instance_id:
        raise Exception(
            "Could not find registered instance with host_id=%s and port=%s"
            % (host_id, port)
        )
    return instance_id


def get_request_ids(request):
    host_id = get_host_id(request.db_session, request.instance.hostname)
    instance_id = get_instance_id(
        request.db_session, host_id, request.instance.pg_port)
    return host_id, instance_id


def parse_start_end(request):
    start = request.handler.get_argument('start', default=None)
    end = request.handler.get_argument('end', default=None)
    try:
        if start:
            start = parse_datetime.parse(start)
        if end:
            end = parse_datetime.parse(end)
    except ValueError:
        raise HTTPError(406, 'Datetime not valid.')

    return start, end


def check_agent_key(session, hostname, pg_data, pg_port, agent_key):
    """Check that the given key matches with the registered one.
    """
    row_key = db.get_agent_key(session, hostname, pg_data, pg_port)

    if not row_key:
        raise Exception("Could not find agent key.")

    if row_key != agent_key:
        raise Exception("Agent key does not match.")


def check_host_key(session, hostname, agent_key):
    """Check that the given key matches with one of .
    """
    matched = False

    if agent_key in db.get_agent_keys(session, hostname):
        matched = True

    if not matched:
        raise Exception("Agent key does not match.")


def insert_metrics(session, host_id, instance_id, data):

    for metric_name in data.keys():
        # Do not try to insert empty lines
        if data[metric_name] is None:
            continue
        if len(data[metric_name]) == 0:
            continue

        # Insert data
        for metric_data in data[metric_name]:
            if metric_name == 'sessions':
                db.insert_metric_sessions(session, instance_id, metric_data)
            elif metric_name == 'xacts':
                db.insert_metric_xacts(session, instance_id, metric_data)
            elif metric_name == 'locks':
                db.insert_metric_locks(session, instance_id, metric_data)
            elif metric_name == 'blocks':
                db.insert_metric_blocks(session, instance_id, metric_data)
            elif metric_name == 'bgwriter':
                db.insert_metric_bgwriter(session, instance_id, metric_data)
            elif metric_name == 'db_size':
                db.insert_metric_db_size(session, instance_id, metric_data)
            elif metric_name == 'tblspc_size':
                db.insert_metric_tblspc_size(session, instance_id, metric_data)
            elif metric_name == 'filesystems_size':
                db.insert_metric_fs_size(session, host_id, metric_data)
            elif metric_name == 'temp_files_size_delta':
                db.insert_metric_temp_files_size_delta(
                    session, instance_id, metric_data
                )
            elif metric_name == 'wal_files':
                db.insert_metric_wal_files(session, instance_id, metric_data)
            elif metric_name == 'cpu':
                db.insert_metric_cpu(session, host_id, metric_data)
            elif metric_name == 'process':
                db.insert_metric_process(session, host_id, metric_data)
            elif metric_name == 'memory':
                db.insert_metric_memory(session, host_id, metric_data)
            elif metric_name == 'loadavg':
                db.insert_metric_loadavg(session, host_id, metric_data)
            elif metric_name == 'vacuum_analyze':
                db.insert_metric_vacuum_analyze(
                    session, instance_id, metric_data
                )
            elif metric_name == 'replication_lag':
                db.insert_metric_replication_lag(
                    session, instance_id, metric_data
                )
            elif metric_name == 'replication_connection':
                db.insert_metric_replication_connection(
                    session, instance_id, metric_data
                )
            elif metric_name == 'heap_bloat':
                db.insert_metric_heap_bloat(session, instance_id, metric_data)
            elif metric_name == 'btree_bloat':
                db.insert_metric_btree_bloat(session, instance_id, metric_data)


def get_instance_checks(session, instance_id):
    # Returns enabled alerting checks as list of tuples:
    # (name, warning threshold, critical threshold)
    checks = session.query(Check).filter(Check.instance_id == instance_id)
    return [(c.name, c.warning, c.critical)
            for c in checks if c.enabled]


def populate_host_checks(session, host_id, instance_id, hostinfo):
    # Populate checks table with bootstraped checks if needed
    for bc in bootstrap_checks(hostinfo):
        # Do not try to add new check if exists
        if session.query(Check).filter(
                Check.host_id == host_id,
                Check.instance_id == instance_id,
                Check.name == bc[0]).count() > 0:
            continue
        c = Check(host_id=host_id,
                  instance_id=instance_id,
                  name=bc[0],
                  enabled=True,
                  warning=bc[1],
                  critical=bc[2],
                  description=check_specs.get(bc[0], {}).get('description'))
        session.add(c)
    session.commit()


def update_collector_status(session, instance_id, status, last_pull=None,
                            last_push=None, last_insert=None):
    cs = CollectorStatus()
    cs.instance_id = instance_id
    cs.status = status
    if last_pull:
        cs.last_pull = last_pull
    if last_push:
        cs.last_push = last_push
    if last_insert:
        cs.last_insert = last_insert

    session.merge(cs)


def build_check_task_options(data, host_id, instance_id, checks, timestamp):
    """Build Task options for check_data_worker worker."""

    return dict(
        host_id=host_id,
        instance_id=instance_id,
        data=preprocess_data(data, checks, timestamp),
    )


def preprocess_data(data, checks, timestamp):
    """Preprocess metrics for alerting."""
    ret = list()

    for check in checks:
        spec = check_specs.get(check[0])
        if not spec:
            continue

        try:
            res = spec.get('preprocess')(data)
        except Exception as e:
            logger.warning(
                "Failed to preprocess alerting check '%s': %s", check[0], e,
            )
            continue

        if not isinstance(res, dict):
            res = {'': res}

        for key, value in res.items():
            ret.append(
                dict(
                    datetime=timestamp,
                    name=check[0],
                    key=key,
                    value=value,
                    warning=check[1],
                    critical=check[2]
                )
            )
    return ret


def check_preprocessed_data(session, host_id, instance_id, ppdata, home):
    # Function in charge of checking preprocessed monitoring values
    keys = dict()

    for raw in ppdata:
        dt = raw.get('datetime')
        name = raw.get('name')
        key = raw.get('key')
        value = raw.get('value')
        warning = raw.get('warning')
        critical = raw.get('critical')

        # Proceed with thresholds comparison
        spec = check_specs.get(name)
        state = 'UNDEF'
        if not spec:
            continue
        if not (spec.get('operator')(value, warning)
                or spec.get('operator')(value, critical)):
            state = 'OK'
        if spec.get('operator')(value, warning):
            state = 'WARNING'
        if spec.get('operator')(value, critical):
            state = 'CRITICAL'

        # Try to find enabled check for this host_id with the same name
        try:
            c = session.query(Check).filter(
                Check.name == unicode(name),
                Check.host_id == host_id,
                Check.instance_id == instance_id,
                Check.enabled == bool(True),
            ).one()
        except NoResultFound:
            continue

        # Update/insert check current state
        try:
            cs = session.query(CheckState).filter(
                CheckState.check_id == c.check_id,
                CheckState.key == unicode(key)
            ).one()
            # State has changed since last time
            if cs.state != state:
                taskmanager.schedule_task(
                    'notify_state_change',
                    listener_addr=os.path.join(home, '.tm.socket'),
                    options={
                        'check_id': c.check_id,
                        'key': key,
                        'value': value,
                        'state': state,
                        'prev_state': cs.state
                    },
                    expire=0,
                )
            cs.state = unicode(state)
            session.merge(cs)

        except NoResultFound:
            cs = CheckState(
                check_id=c.check_id, key=unicode(key), state=unicode(state)
            )
            session.add(cs)

        session.flush()
        # Append state change if any to history
        db.append_state_changes(
            session,
            dt,
            c.check_id,
            cs.state,
            cs.key,
            value,
            warning,
            critical,
        )

        if c.check_id not in keys:
            keys[c.check_id] = list()
        keys[c.check_id].append(cs.key)

        session.commit()
        session.expunge_all()

    # Purge CheckState
    for check_id, ks in keys.items():
        db.purge_check_states(session, check_id, ks)
        session.commit()

    # Get the list of check_id for the given instance
    req = session.query(Check).filter(Check.instance_id == instance_id)
    all_check_ids = [check.check_id for check in req]

    # Set to UNDEF each unchecked check for the given instance
    # This may happen when postgres is not available
    db.undef_check_states(session, all_check_ids, keys.keys())
    session.commit()



import httplib
import urllib2
import ssl
import socket
import json


class VerifiedHTTPSConnection(httplib.HTTPSConnection):
    def connect(self):
        sock = socket.create_connection((self.host, self.port), self.timeout)
        if self._tunnel_host:
            self.sock = sock
            self._tunnel()
        self.sock = ssl.wrap_socket(
            sock,
            self.key_file,
            self.cert_file,
            cert_reqs=ssl.CERT_REQUIRED,
            ca_certs=CA_CERT_FILE)


class VerifiedHTTPSHandler(urllib2.HTTPSHandler):
    def __init__(self, connection_class=VerifiedHTTPSConnection):
        self.specialized_conn_class = connection_class
        urllib2.HTTPSHandler.__init__(self)

    def https_open(self, req):
        return self.do_open(self.specialized_conn_class, req)


class UnverifiedHTTPSConnection(httplib.HTTPSConnection):
    # HTTPS connection class, without any SSL cert. check.
    def connect(self):
        sock = socket.create_connection((self.host, self.port), self.timeout)
        if self._tunnel_host:
            self.sock = sock
            self._tunnel()
        self.sock = ssl.wrap_socket(sock)


class UnverifiedHTTPSHandler(urllib2.HTTPSHandler):
    # HTTPS connection handler
    def __init__(self, connection_class=UnverifiedHTTPSConnection):
        self.specialized_conn_class = connection_class
        urllib2.HTTPSHandler.__init__(self)

    def https_open(self, req):
        return self.do_open(self.specialized_conn_class, req)


class RequestWithMethod(urllib2.Request):
    def __init__(self, *args, **kwargs):
        self._method = kwargs.pop('method', None)
        urllib2.Request.__init__(self, *args, **kwargs)

    def get_method(self):
        return self._method if self._method else super(RequestWithMethod,
                                                       self).get_method()


def temboard_request(in_ca_cert_file, method, url, headers=None, data=None):
    global CA_CERT_FILE
    CA_CERT_FILE = in_ca_cert_file
    if in_ca_cert_file is None:
        https_handler = UnverifiedHTTPSHandler()
    else:
        https_handler = VerifiedHTTPSHandler()
    url_opener = urllib2.build_opener(https_handler)
    headers_list = []
    for key, val in headers.iteritems():
        headers_list.append((key, val))
    url_opener.addheaders = headers_list
    if data is not None:
        request = RequestWithMethod(url, data=json.dumps(data), method=method)
    else:
        request = RequestWithMethod(url, method=method)
    handle = url_opener.open(request)
    response = handle.read()
    handle.close()
    return response


class TemboardError(Exception):
    def __init__(self, code, message):
        Exception.__init__(self, message)
        self.code = code
        self.message = message


def temboard_discover(in_ca_cert_file, hostname, port):
    try:
        res = temboard_request(
            in_ca_cert_file,
            method='GET',
            url='https://%s:%s/discover' % (hostname, port),
            headers={"Content-type": "application/json"})
        return json.loads(res)
    except urllib2.HTTPError as e:
        raise TemboardError(e.code, json.loads(e.read())['error'])
    except Exception as e:
        raise TemboardError(500, str(e))


def temboard_profile(in_ca_cert_file, hostname, port, xsession):
    try:
        res = temboard_request(
            in_ca_cert_file,
            method='GET',
            url='https://%s:%s/profile' % (hostname, port),
            headers={
                "Content-type": "application/json",
                "X-Session": xsession
            })
        return json.loads(res)
    except urllib2.HTTPError as e:
        raise TemboardError(e.code, json.loads(e.read())['error'])
    except Exception as e:
        raise TemboardError(500, str(e))



# coding: utf-8

import imp
import logging.config
import os
import pkg_resources
import socket
import sys
from argparse import (
    ArgumentParser,
    SUPPRESS as UNDEFINED_ARGUMENT,
    _VersionAction,
)
from concurrent.futures import ThreadPoolExecutor

import tornado.ioloop
import tornado.web
from tornado import autoreload

from .model import configure as configure_db_session, check_schema
from .toolkit import taskmanager
from .toolkit import validators as v
from .toolkit.app import define_core_arguments
from .toolkit.configuration import OptionSpec
from .toolkit.errors import UserError
from .toolkit.proctitle import ProcTitleManager
from .toolkit.services import (
    Service,
    ServicesManager,
)
from .toolkit.tasklist.sqlite3_engine import TaskListSQLite3Engine
from .web import Error404Handler, app

from .daemon import daemonize
from .pluginsmgmt import load_plugins
from .autossl import AutoHTTPSServer
from .toolkit.app import BaseApplication
from .version import __version__, format_version, inspect_versions


logger = logging.getLogger('temboardui')


def legacy_bootstrap(config):
    # Compat with legacy load_plugins
    config.plugins = {}

    autoreload.watch(config.temboard.configfile)

    # Run temboard as a background daemon.
    if config.temboard.daemonize:
        daemonize(config.temboard.pidfile, config)

    return config


def legacy_enable_plugins(self, plugin_names):
    # Load and enable legacy plugins in tornado app.
    plugins = load_plugins(plugin_names, self.config)
    plugins_conf = dict()
    self.workersets = []
    self.loaded_plugins = []
    for key, val in plugins.iteritems():
        self.add_rules(val['routes'])
        plugins_conf[key] = val['configuration']
        if val['workers']:
            self.workersets.append(val['workers'])
        if key not in self.loaded_plugins:
            self.loaded_plugins.append(key)
    return plugins_conf


def bootstrap_tornado_app(app, config):
    app.config = config

    base_path = os.path.dirname(__file__)
    # Load handlers
    __import__('temboardui.handlers.home')
    __import__('temboardui.handlers.notification')
    __import__('temboardui.handlers.settings.group')
    __import__('temboardui.handlers.settings.instance')
    __import__('temboardui.handlers.settings.user')
    __import__('temboardui.handlers.settings.notifications')
    __import__('temboardui.handlers.user')

    app.configure(
        cookie_secret=config.temboard['cookie_secret'],
        debug=config.logging.debug,
        template_path=base_path + "/templates",
        default_handler_class=Error404Handler,
    )

    return app


def finalize_tornado_app(app, config, plugins):
    base_path = os.path.dirname(__file__)
    handlers = [
        (r"/css/(.*)", tornado.web.StaticFileHandler, {
            'path': base_path + '/static/css'
        }),
        (r"/js/(.*)", tornado.web.StaticFileHandler, {
            'path': base_path + '/static/js'
        }),
        (r"/images/(.*)", tornado.web.StaticFileHandler, {
            'path': base_path + '/static/images'
        }),
        (r"/fonts/(.*)", tornado.web.StaticFileHandler, {
            'path': base_path + '/static/fonts'
        })
    ]

    config.plugins = legacy_enable_plugins(app, plugins)
    # Append rules *after* plugins because plugins shares same namespace for
    # static rules, i.e. /js/.* is a fallback for /js/dashboard/.*.
    app.add_rules(handlers)


def map_pgvars(environ):
    pgvar_map = dict(
        PGHOST='TEMBOARD_REPOSITORY_HOST',
        PGPORT='TEMBOARD_REPOSITORY_PORT',
        PGUSER='TEMBOARD_REPOSITORY_USER',
        PGPASSWORD='TEMBOARD_REPOSITORY_PASSWORD',
        PGDATABASE='TEMBOARD_REPOSITORY_DBNAME',
    )
    mapped = environ.copy()
    for pgvar, tbvar in pgvar_map.items():
        if tbvar in environ:
            continue

        try:
            mapped[tbvar] = environ[pgvar]
            logger.debug("Read %s from environ.", pgvar)
        except KeyError:
            pass

    return mapped


class SchedulerService(taskmanager.SchedulerService):
    def apply_config(self):
        super(SchedulerService, self).apply_config()
        if self.scheduler:
            # Set legacy config context.
            self.scheduler.set_context(
                'config',
                {
                    # Wrap settings in dict for JSON serializable.
                    'plugins': dict(self.app.config.plugins),
                    'temboard': dict(self.app.config.temboard),
                    'repository': dict(self.app.config.repository),
                }
            )


class TornadoService(Service):
    def setup(self):
        config = self.app.config
        ssl_ctx = {
            'certfile': config.temboard.ssl_cert_file,
            'keyfile': config.temboard.ssl_key_file,
        }
        server = AutoHTTPSServer(self.app.webapp, ssl_options=ssl_ctx)
        try:
            server.listen(
                config.temboard.port, address=config.temboard.address)
        except socket.error as e:
            logger.error("FATAL: " + str(e) + '. Quit')
            sys.exit(3)

    def serve(self):
        with self:
            # Automatically reload modified modules (from Tornado's
            # Application.__init__). This code must be done here *after*
            # daemonize, because it instanciates ioloop for current PID.
            if self.app.webapp.settings.get('autoreload'):
                autoreload.start()

            logger.info(
                "Serving temboardui on https://%s:%d",
                self.app.config.temboard.address,
                self.app.config.temboard.port)
            tornado.ioloop.IOLoop.instance().start()


class VersionAction(_VersionAction):
    def __call__(self, parser, *_):
        print(format_version().strip())
        parser.exit()


def define_arguments(parser):
    define_core_arguments(parser)
    parser.add_argument(
        '-V', '--version',
        action=VersionAction,
        help='show version and exit',
    )
    parser.add_argument(
        '-d', '--daemon',
        action='store_true', dest='temboard_daemonize',
        help="Run in background.",
    )
    parser.add_argument(
        '-p', '--pid-file',
        action='store', dest='temboard_pidfile',
        help="PID file.", metavar='PIDFILE',
    )


def cookie_secret(raw):
    length = len(raw)
    if length < 10:
        raise ValueError("cookie secret is shorter than 10 chars.")
    if length > 128:
        raise ValueError("cookie secret is longer than 128 chars.")
    return raw


def list_options_specs():
    s = 'temboard'
    # Manage plugin list here until we use plugin entrypoint.
    yield OptionSpec(
        s, 'plugins',
        default=TemboardApplication.DEFAULT_PLUGINS,
        validator=v.jsonlist,
    )
    yield OptionSpec(s, 'daemonize', default=False)
    yield OptionSpec(s, 'pidfile', default='/run/temboard.pid')
    yield OptionSpec(s, 'address', default='0.0.0.0', validator=v.address)
    yield OptionSpec(s, 'port', validator=v.port, default=8888)
    yield OptionSpec(
        s, 'ssl_cert_file',
        default=OptionSpec.REQUIRED, validator=v.file_)
    yield OptionSpec(
        s, 'ssl_key_file',
        default=OptionSpec.REQUIRED, validator=v.file_)
    yield OptionSpec(s, 'ssl_ca_cert_file', validator=v.file_)
    yield OptionSpec(s, 'cookie_secret', validator=cookie_secret)
    home = os.environ.get('HOME', '/var/lib/temboard')
    yield OptionSpec(s, 'home', default=home, validator=v.writeabledir)

    s = 'repository'
    yield OptionSpec(s, 'host', default='/var/run/postgresql')
    yield OptionSpec(s, 'instance', default='main')
    yield OptionSpec(s, 'port', default=5432, validator=v.port)
    yield OptionSpec(s, 'user', default='temboard')
    yield OptionSpec(s, 'password', default='temboard')
    yield OptionSpec(s, 'dbname', default='temboard')

    s = 'notifications'
    yield OptionSpec(s, 'smtp_host', default=None)
    yield OptionSpec(s, 'smtp_port', default=None, validator=v.port)
    yield OptionSpec(s, 'smtp_tls', default=False, validator=v.boolean)
    yield OptionSpec(s, 'smtp_login', default=None)
    yield OptionSpec(s, 'smtp_password', default=None)
    yield OptionSpec(s, 'smtp_from_addr', default=None)
    yield OptionSpec(s, 'twilio_account_sid', default=None)
    yield OptionSpec(s, 'twilio_auth_token', default=None)
    yield OptionSpec(s, 'twilio_from', default=None)

    s = 'monitoring'
    yield OptionSpec(s, 'purge_after', default=None, validator=v.nday)

    s = 'statements'
    yield OptionSpec(s, 'purge_after', default=7, validator=v.nday)


class TemboardApplication(BaseApplication):
    DEFAULT_CONFIGFILES = [
        '/etc/temboard/temboard.conf',
        'temboard.conf',
    ]
    DEFAULT_PLUGINS = [
        'activity',
        'dashboard',
        'monitoring',
        'pgconf',
        'maintenance',
        'statements',
    ]
    PROGRAM = 'temboard'
    REPORT_URL = "https://github.com/dalibo/temboard/issues"
    VERSION = __version__

    def filter_plugins(self):
        # Returns two lists of plugin names from config.temboard.plugins, the
        # first list is legacy plugins, the second is external plugins.

        legacy_plugins = []
        ep_plugins = []

        path = os.path.dirname(__file__) + '/plugins'

        logger.debug("Looping over all plugins.")
        for name in self.config.temboard.plugins:
            if any(pkg_resources.iter_entry_points(self.with_plugins, name)):
                logger.debug("%s is a new plugin.", name)
                ep_plugins.append(name)
            else:
                try:
                    fp, _, _ = imp.find_module(name, [path])
                except ImportError:
                    raise UserError("Missing plugin: %s." % name)
                else:
                    legacy_plugins.append(name)
                    logger.debug("%s is a legacy plugin.", name)

        return legacy_plugins, ep_plugins

    def apply_config(self):
        bootstrap_tornado = not hasattr(self, 'webapp')
        if bootstrap_tornado:
            # For now, just create web app once. One time, we'll be able to
            # unload plugin routes.
            self.webapp = bootstrap_tornado_app(app, self.config)
            self.webapp.executor = ThreadPoolExecutor(12)
            self.webapp.temboard_app = self

        legacy_plugins, ep_plugins = self.filter_plugins()
        self.config.temboard.plugins = ep_plugins

        super(TemboardApplication, self).apply_config()

        finalize_tornado_app(app, self.config, plugins=legacy_plugins)

        self.webapp.engine = configure_db_session(self.config.repository)

    def main(self, argv, environ):

        # C O N F I G U R A T I O N

        parser = ArgumentParser(
            prog='temboard',
            description="temBoard web UI.",
            argument_default=UNDEFINED_ARGUMENT,
        )
        define_arguments(parser)
        args = parser.parse_args(argv)
        environ = map_pgvars(environ)
        self.bootstrap(args=args, environ=environ)

        setproctitle = ProcTitleManager(prefix='temboard: ')
        setproctitle.setup()

        versions = inspect_versions()
        logger.info(
            "Running on %s %s.",
            versions['distname'], versions['distversion'])
        logger.info(
            "Using Python %s (%s).",
            versions['python'], versions['pythonbin'])
        logger.info(
            "Using Psycopg2 %s, Tornado %s and SQLAlchemy %s",
            versions['psycopg2'], versions['tornado'], versions['sqlalchemy'],
        )
        logging.getLogger('alembic').setLevel(logging.WARN)
        # Manage logging_debug default until we use toolkit OptionSpec.
        legacy_bootstrap(self.config)

        services = ServicesManager()

        # T A S K   M A N A G E R

        task_queue = taskmanager.Queue()
        event_queue = taskmanager.Queue()

        self.worker_pool = taskmanager.WorkerPoolService(
            app=self, name=u'worker pool',
            task_queue=task_queue, event_queue=event_queue,
            setproctitle=setproctitle,
        )
        self.scheduler = SchedulerService(
            app=self, name=u'scheduler',
            task_queue=task_queue, event_queue=event_queue,
            setproctitle=setproctitle,
        )

        # TaskList engine setup must be done before we load the plugins
        self.scheduler.task_list_engine = TaskListSQLite3Engine(
            os.path.join(self.config.temboard['home'], 'server_tasks.db')
        )

        self.apply_config()

        check_schema(self.config)

        self.worker_pool.apply_config()
        services.add(self.worker_pool)

        for wset in self.webapp.workersets:
            self.worker_pool.add(wset)

        self.scheduler.apply_config()
        services.add(self.scheduler)

        # H T T P   S E R V E R

        webservice = TornadoService(
            app=self, name=u'web', services=services,
            setproctitle=setproctitle,
        )
        with services:
            webservice.run()


main = TemboardApplication(
    specs=list_options_specs(),
    with_plugins="temboard.plugins",
)


if __name__ == "__main__":
    sys.exit(main())



import logging
from os.path import realpath

import tornado.web
from tornado.escape import json_encode

from temboardui.temboardclient import TemboardError
from temboardui.web import (
    Blueprint,
    TemplateRenderer,
)


blueprint = Blueprint()
blueprint.generic_proxy(r"/dashboard")
logger = logging.getLogger(__name__)
plugin_path = realpath(__file__ + '/..')
render_template = TemplateRenderer(plugin_path + '/templates')

PLUGIN_NAME = 'dashboard'


def configuration(config):
    return {}


def get_routes(config):
    routes = blueprint.rules + [
        (r"/js/dashboard/(.*)", tornado.web.StaticFileHandler, {
            'path': plugin_path + "/static/js"
        }),
    ]
    return routes


@blueprint.instance_route(r"/dashboard")
def dashboard(request):
    request.instance.check_active_plugin(PLUGIN_NAME)

    try:
        agent_username = request.instance.get_profile()['username']
    except Exception:
        agent_username = None

    try:
        config = request.instance.get('/dashboard/config')
    except TemboardError as e:
        if 404 != e.code:
            raise
        logger.debug("Fallback dashboard config.")
        config = {
            'history_length': 150,
            'scheduler_interval': 2
        }

    history = request.instance.get('/dashboard/history')
    if history:
        last_data = history[-1]
    else:
        last_data = request.instance.get('/dashboard/live')

    return render_template(
        'dashboard.html',
        nav=True, role=request.current_user,
        instance=request.instance,
        agent_username=agent_username,
        plugin=PLUGIN_NAME,
        config=json_encode(config),
        dashboard=last_data,
        history=json_encode(history or ''),
    )



from os import path
import tornado.web

from temboardui.web import (
    Blueprint,
    TemplateRenderer,
)


PLUGIN_NAME = 'maintenance'
blueprint = Blueprint()
blueprint.generic_proxy(r'/maintenance/.*/schema/.*/table/.*/'
                        '(?:vacuum|analyze)',
                        methods=['POST'])
blueprint.generic_proxy(r'/maintenance/.*/(?:vacuum|analyze)',
                        methods=['POST'])
blueprint.generic_proxy(r'/maintenance/.*/schema/.*/(?:index|table)/.*/'
                        '(?:reindex)',
                        methods=['POST'])
blueprint.generic_proxy(r'/maintenance/.*/(?:reindex)',
                        methods=['POST'])
blueprint.generic_proxy(r'/maintenance/reindex/.*',
                        methods=['DELETE'])
blueprint.generic_proxy(r'/maintenance/(?:vacuum|analyze)/.*',
                        methods=['DELETE'])
blueprint.generic_proxy(r'/maintenance/.*')
blueprint.generic_proxy(r'/maintenance')
plugin_path = path.dirname(path.realpath(__file__))
render_template = TemplateRenderer(plugin_path + '/templates')


def configuration(config):
    return {}


def get_routes(config):
    routes = blueprint.rules + [
        (r"/js/maintenance/(.*)", tornado.web.StaticFileHandler, {
            'path': plugin_path + "/static/js"
        }),
    ]
    return routes


def get_agent_username(request):
    try:
        agent_username = request.instance.get_profile()['username']
    except Exception:
        agent_username = None
    return agent_username


@blueprint.instance_route(r'/maintenance')
def maintenance(request):
    request.instance.check_active_plugin(PLUGIN_NAME)
    return render_template(
        'index.html',
        nav=True,
        agent_username=get_agent_username(request),
        instance=request.instance,
        plugin=PLUGIN_NAME,
        role=request.current_user,
    )


@blueprint.instance_route(r'/maintenance/(.*)/schema/(.*)/table/(.*)')
def table(request, database, schema, table):
    request.instance.check_active_plugin(PLUGIN_NAME)
    agent_username = get_agent_username(request)
    xsession = request.instance.xsession if agent_username else None
    return render_template(
        'table.html',
        nav=True,
        agent_username=agent_username,
        instance=request.instance,
        plugin=PLUGIN_NAME,
        xsession=xsession,
        role=request.current_user,
        database=database,
        schema=schema,
        table=table,
    )


@blueprint.instance_route(r'/maintenance/(.*)/schema/(.*)')
def schema(request, database, schema):
    request.instance.check_active_plugin(PLUGIN_NAME)
    agent_username = get_agent_username(request)
    xsession = request.instance.xsession if agent_username else None
    return render_template(
        'schema.html',
        nav=True,
        agent_username=agent_username,
        instance=request.instance,
        plugin=PLUGIN_NAME,
        xsession=xsession,
        role=request.current_user,
        database=database,
        schema=schema,
    )


@blueprint.instance_route(r'/maintenance/(.*)')
def database(request, database):
    request.instance.check_active_plugin(PLUGIN_NAME)
    agent_username = get_agent_username(request)
    xsession = request.instance.xsession if agent_username else None
    return render_template(
        'database.html',
        nav=True,
        agent_username=get_agent_username(request),
        instance=request.instance,
        plugin=PLUGIN_NAME,
        xsession=xsession,
        role=request.current_user,
        database=database,
    )



import logging
import cStringIO
from datetime import datetime
from textwrap import dedent

import tornado.web
import tornado.escape

from temboardui.plugins.monitoring.model.orm import (
    Check,
    CheckState,
)
from temboardui.web import (
    HTTPError,
    jsonify,
)

from . import (
    blueprint,
    render_template,
)
from ..tools import get_request_ids, parse_start_end
from ..alerting import checks_info, check_state_detail, check_specs

logger = logging.getLogger(__name__)


def sql_json_query(request, query, *args):
    # Helper to query JSON output from PostgreSQL.

    cur = request.db_session.connection().connection.cursor()
    query = cur.mogrify(query, args)
    data_buffer = cStringIO.StringIO()
    cur.copy_expert(query, data_buffer)
    cur.close()
    data = data_buffer.getvalue()
    data_buffer.close()
    try:
        return tornado.escape.json_decode(data)
    except Exception as e:
        logger.error("Failed to parse JSON from Postgres: %s", e)
        logger.error("Postgres output is: %r", data)
        return []


@blueprint.instance_route(r"/alerting/alerts.json")
def alerts(request):
    host_id, instance_id = get_request_ids(request)

    query = dedent("""\
    COPY (
        SELECT array_to_json(coalesce(array_agg(x), '{}'))
        FROM (
            SELECT json_build_object('description', c.description, 'name', c.name, 'key', sc.key, 'state', sc.state, 'datetime', sc.datetime, 'value', sc.value, 'warning', sc.warning, 'critical', sc.critical) as x
            FROM monitoring.state_changes sc JOIN monitoring.checks c ON (sc.check_id = c.check_id)
            WHERE c.host_id = %s
              AND c.instance_id = %s
              AND (sc.state = 'WARNING' OR sc.state = 'CRITICAL')
            ORDER BY sc.datetime desc
            LIMIT 20
        ) as tab
    ) TO STDOUT
    """)  # noqa

    # Tornado refuses to send lists as JSON. We must explicitly use jsonify.
    # Cf. https://github.com/tornadoweb/tornado/issues/1009
    return jsonify(sql_json_query(request, query, host_id, instance_id))


@blueprint.instance_route(r"/alerting")
def index(request):
    try:
        agent_username = request.instance.get_profile()['username']
    except Exception:
        # Monitoring plugin doesn't require agent authentication since we
        # already have the data.
        # Don't fail if there's a session error (for example when the agent
        # has been restarted)
        agent_username = None

    return render_template(
        'alerting.checks.html',
        nav=True, role=request.current_user,
        instance=request.instance, agent_username=agent_username,
        plugin='alerting',  # we cheat here
    )


@blueprint.instance_route("/alerting/checks.json", methods=['GET', 'POST'])
def checks(request):
    host_id, instance_id = get_request_ids(request)
    if 'GET' == request.method:
        data = checks_info(request.db_session, host_id, instance_id)
        for datum in data:
            spec = check_specs[datum['name']]
            if 'value_type' in spec:
                datum['value_type'] = spec['value_type']
        return jsonify(data)
    else:
        post = tornado.escape.json_decode(request.body)
        if 'checks' not in post or type(post.get('checks')) is not list:
            raise HTTPError(400, "Post data not valid.")

        for row in post['checks']:
            if row.get('name') not in check_specs:
                raise HTTPError(404, "Unknown check '%s'" % row.get('name'))

        for row in post['checks']:
            # Find the check from its name
            check = request.db_session.query(Check).filter(
                        Check.name == unicode(row.get('name')),
                        Check.host_id == host_id,
                        Check.instance_id == instance_id).first()
            enabled_before = check.enabled

            if u'enabled' in row:
                enabled_after = bool(row.get(u'enabled'))
                check.enabled = enabled_after
                # detect any change from enabled to disabled
                is_getting_disabled = enabled_before and not enabled_after
            if u'warning' in row:
                warning = row.get(u'warning')
                if type(warning) not in (int, float):
                    raise HTTPError(400, "Post data not valid.")
                check.warning = warning
            if u'critical' in row:
                critical = row.get(u'critical')
                if type(critical) not in (int, float):
                    raise HTTPError(400, "Post data not valid.")
                check.critical = critical
            if u'description' in row:
                check.description = row.get(u'description')

            request.db_session.merge(check)

            if is_getting_disabled:
                cs = request.db_session.query(CheckState).filter(
                    CheckState.check_id == check.check_id,
                )
                for i in cs:
                    i.state = unicode('UNDEF')
                    request.db_session.merge(i)
                    request.db_session.execute(
                        "SELECT monitoring.append_state_changes(:d, :i,"
                        ":s, :k, :v, :w, :c)",
                        {'d': datetime.utcnow(), 'i': check.check_id,
                         's': 'UNDEF', 'k': i.key, 'v': None,
                         'w': check.warning, 'c': check.critical})

        request.db_session.commit()

        return {}


@blueprint.instance_route(r"/alerting/([a-z\-_.0-9]{1,64})")
def check(request, name):
    try:
        agent_username = request.instance.get_profile()['username']
    except Exception:
        # Monitoring plugin doesn't require agent authentication since we
        # already have the data.
        # Don't fail if there's a session error (for example when the agent
        # has been restarted)
        agent_username = None

    host_id, instance_id = get_request_ids(request)
    query = dedent("""\
    SELECT *
    FROM monitoring.checks
    WHERE host_id = :host_id
      AND instance_id = :instance_id
      AND name = :check_name
    """)
    res = request.db_session.execute(query, dict(
        host_id=host_id,
        instance_id=instance_id,
        check_name=name,
    ))
    check = res.fetchone()
    spec = check_specs[name]

    return render_template(
        'alerting.check.html',
        nav=True, role=request.current_user,
        instance=request.instance, agent_username=agent_username,
        plugin='alerting',  # we cheat here
        check=check,
        value_type=spec.get('value_type'),
    )


@blueprint.instance_route(r"/alerting/check_changes/([a-z\-_.0-9]{1,64}).json")
def check_changes(request, name):
    host_id, instance_id = get_request_ids(request)
    start, end = parse_start_end(request)

    query = dedent("""\
    COPY (
        SELECT array_to_json(coalesce(array_agg(json_build_object(
            'datetime', f.datetime,
            'enabled', f.enabled,
            'warning', f.warning,
            'critical', f.critical,
            'description', f.description
        )), '{}')) FROM monitoring.get_check_changes(%s, %s, %s, %s, %s) f
    ) TO STDOUT
    """)
    return jsonify(sql_json_query(
        request, query,
        host_id, instance_id, name, start, end,
    ))


@blueprint.instance_route(r"/alerting/state_changes/([a-z\-_.0-9]{1,64}).json")
def state_changes(request, name):
    host_id, instance_id = get_request_ids(request)
    start, end = parse_start_end(request)
    key = request.handler.get_argument('key', default=None)
    if name not in check_specs:
        raise HTTPError(404, "Unknown check '%s'" % name)

    query = dedent("""\
    COPY (
        SELECT array_to_json(coalesce(array_agg(json_build_object(
            'datetime', f.datetime,
            'state', f.state,
            'value', f.value,
            'warning', f.warning,
            'critical', f.critical
        )), '{}')) FROM monitoring.get_state_changes(%s, %s, %s, %s, %s, %s) f
    ) TO STDOUT
    """)

    return jsonify(sql_json_query(
        request, query,
        host_id, instance_id, name, key, start, end,
    ))


@blueprint.instance_route(r"/alerting/states/([a-z\-_.0-9]{1,64}).json")
def states(request, name):
    host_id, instance_id = get_request_ids(request)
    if name not in check_specs:
        raise HTTPError(404, "Unknown check '%s'" % name)

    detail = check_state_detail(request.db_session, host_id, instance_id, name)
    for d in detail:
        spec = check_specs[name]
        if 'value_type' in spec:
            d['value_type'] = spec['value_type']

    return jsonify(detail)



import logging
import sys
from argparse import (
    ArgumentParser,
    SUPPRESS as UNDEFINED_ARGUMENT,
)

import alembic.command
import alembic.config
import sqlalchemy.exc

from .__main__ import VersionAction, map_pgvars
from .model import build_alembic_config, check_schema
from .toolkit import validators as v
from .toolkit.app import (
    BaseApplication,
    define_core_arguments,
)
from .toolkit.configuration import OptionSpec
from .toolkit.errors import UserError
from .version import __version__, inspect_versions


logger = logging.getLogger(__name__)


class MigrateDBApplication(BaseApplication):
    DEFAULT_CONFIGFILES = [
        '/etc/temboard/temboard.conf',
        'temboard.conf',
    ]
    PROGRAM = 'temboard-migratedb'
    REPORT_URL = "https://github.com/dalibo/temboard/issues"
    VERSION = __version__

    def main(self, argv, environ):
        parser = ArgumentParser(
            prog='temboard-migratedb',
            description="temBoard schema migrator.",
            argument_default=UNDEFINED_ARGUMENT,
        )
        define_arguments(parser)
        args = parser.parse_args(argv)
        environ = map_pgvars(environ)
        self.bootstrap(args=args, environ=environ)

        versions = inspect_versions()
        logger.debug(
            "Using Python %s (%s).",
            versions['python'], versions['pythonbin'])
        logger.debug(
            "Using Psycopg2 %s, Alembic %s and SQLAlchemy %s",
            versions['psycopg2'], versions['alembic'], versions['sqlalchemy'],
        )

        command_method = 'command_' + args.command
        try:
            getattr(self, command_method)(args)
        except sqlalchemy.exc.OperationalError as e:
            raise UserError("Failed to query Postgres server: %s." % e)

    def command_check(self, args):
        logging.getLogger('alembic').setLevel(logging.WARN)
        check_schema(self.config)

    def command_stamp(self, args):
        logging.getLogger('alembic').setLevel(logging.WARN)
        alembic_cfg = build_alembic_config(self.config)
        alembic.command.stamp(alembic_cfg, 'head')
        logger.info("Database marked as up to date.")

    def command_upgrade(self, args):
        alembic_cfg = build_alembic_config(self.config)
        alembic.command.upgrade(alembic_cfg, 'head')
        logger.info("Database up to date.")


def define_arguments(parser):
    define_core_arguments(parser)
    parser.add_argument(
        '-V', '--version',
        action=VersionAction,
        help='show version and exit',
    )
    sub = parser.add_subparsers(
        metavar="COMMAND",
        dest="command",
        help="Operation to execute on temBoard database.")
    sub.add_parser(
        "check",
        help='Check schema synchronisation status only.'
    )
    sub.add_parser(
        "stamp",
        help="Mark database as uptodate without migrating.",
    )
    sub.add_parser(
        "upgrade",
        help="Upgrade temBoard database to latest revision.",
    )


def list_options_specs():
    s = 'repository'
    yield OptionSpec(s, 'host', default='/var/run/postgresql')
    yield OptionSpec(s, 'instance', default='main')
    yield OptionSpec(s, 'port', default=5432, validator=v.port)
    yield OptionSpec(s, 'user', default='temboard')
    yield OptionSpec(s, 'password', default='temboard')
    yield OptionSpec(s, 'dbname', default='temboard')


main = MigrateDBApplication(specs=list_options_specs(), with_plugins=None)


if __name__ == "__main__":
    sys.exit(main())



from sqlalchemy.sql import text

from sqlalchemy.schema import (
    MetaData,
    Table,
    Column,
    CheckConstraint,
    ForeignKeyConstraint
)
from sqlalchemy.types import (
    Integer,
    UnicodeText,
    BigInteger,
    Boolean,
)

metadata = MetaData()

roles = Table(
    'roles', metadata,
    Column('role_name', UnicodeText, primary_key=True),
    Column('role_password', UnicodeText, nullable=False),
    Column('role_email', UnicodeText, nullable=False),
    Column('role_phone', UnicodeText),
    Column('is_active', Boolean, nullable=False, server_default=text('True')),
    Column('is_admin', Boolean, nullable=False, server_default=text('False')),
    schema="application")

groups = Table(
    'groups', metadata,
    Column('group_name', UnicodeText, nullable=False, primary_key=True),
    Column('group_description', UnicodeText),
    Column('group_kind', UnicodeText, nullable=False, primary_key=True),
    CheckConstraint("group_kind IN ('instance', 'role')"),
    schema="application")

instances = Table(
    'instances', metadata,
    Column('agent_address', UnicodeText, nullable=False, primary_key=True),
    Column('agent_port', Integer, nullable=False, primary_key=True),
    Column('agent_key', UnicodeText),
    Column('hostname', UnicodeText, nullable=False),
    Column('cpu', Integer),
    Column('memory_size', BigInteger),
    Column('pg_port', Integer),
    Column('pg_version', UnicodeText),
    Column('pg_version_summary', UnicodeText),
    Column('pg_data', UnicodeText),
    Column('notify', Boolean, nullable=False, server_default=text('True')),
    Column('comment', UnicodeText),
    schema="application")

plugins = Table(
    'plugins', metadata,
    Column('agent_address', UnicodeText, nullable=False, primary_key=True),
    Column('agent_port', Integer, nullable=False, primary_key=True),
    Column('plugin_name', UnicodeText, nullable=False, primary_key=True),
    ForeignKeyConstraint(['agent_address', 'agent_port'],
                         ['application.instances.agent_address',
                          'application.instances.agent_port'],
                         ondelete="CASCADE", onupdate="CASCADE"),
    schema="application")

instance_groups = Table(
    'instance_groups', metadata,
    Column('agent_address', UnicodeText, nullable=False, primary_key=True),
    Column('agent_port', Integer, nullable=False, primary_key=True),
    Column('group_name', UnicodeText, nullable=False, primary_key=True),
    Column('group_kind', UnicodeText, nullable=False,
           server_default=text('instance')),
    CheckConstraint("group_kind = 'instance'"),
    ForeignKeyConstraint(['agent_address', 'agent_port'],
                         ['application.instances.agent_address',
                          'application.instances.agent_port'],
                         ondelete="CASCADE", onupdate="CASCADE"),
    ForeignKeyConstraint(['group_name', 'group_kind'],
                         ['application.groups.group_name',
                          'application.groups.group_kind'],
                         ondelete="CASCADE", onupdate="CASCADE"),
    schema="application")

role_groups = Table(
    'role_groups', metadata,
    Column('role_name', UnicodeText, nullable=False, primary_key=True),
    Column('group_name', UnicodeText, nullable=False, primary_key=True),
    Column('group_kind', UnicodeText, nullable=False,
           server_default=text('role')),
    CheckConstraint("group_kind = 'role'"),
    ForeignKeyConstraint(['role_name'],
                         ['application.roles.role_name'],
                         ondelete="CASCADE", onupdate="CASCADE"),
    ForeignKeyConstraint(['group_name', 'group_kind'],
                         ['application.groups.group_name',
                          'application.groups.group_kind'],
                         ondelete="CASCADE", onupdate="CASCADE"),
    schema="application")

access_role_instance = Table(
    'access_role_instance', metadata,
    Column('role_group_name', UnicodeText, nullable=False, primary_key=True),
    Column('role_group_kind', UnicodeText, nullable=False,
           server_default=text('role')),
    Column('instance_group_name', UnicodeText, nullable=False,
           primary_key=True),
    Column('instance_group_kind', UnicodeText, nullable=False,
           server_default=text('instance')),
    CheckConstraint("role_group_kind = 'role'"),
    CheckConstraint("instance_group_kind = 'instance'"),
    ForeignKeyConstraint(['role_group_name', 'role_group_kind'],
                         ['application.groups.group_name',
                          'application.groups.group_kind'],
                         ondelete="CASCADE", onupdate="CASCADE"),
    ForeignKeyConstraint(['instance_group_name', 'instance_group_kind'],
                         ['application.groups.group_name',
                          'application.groups.group_kind'],
                         ondelete="CASCADE", onupdate="CASCADE"),
    schema="application")



"""
Mapping of the database for use with SQLAlchemy.
"""

from sqlalchemy.sql import text

from sqlalchemy.schema import (
    MetaData,
    Table,
    Column,
    ForeignKey,
    UniqueConstraint,
)
from sqlalchemy.types import (
    BigInteger,
    Boolean,
    DateTime,
    Integer,
    REAL,
    UnicodeText,
)

metadata = MetaData()

hosts = Table(
    'hosts', metadata,
    Column('host_id', Integer, primary_key=True),
    Column('hostname', UnicodeText, nullable=False, unique=True),
    Column('os', UnicodeText, nullable=False),
    Column('os_version', UnicodeText, nullable=False),
    Column('os_flavour', UnicodeText),
    Column('cpu_count', Integer),
    Column('cpu_arch', UnicodeText),
    Column('memory_size', BigInteger),
    Column('swap_size', BigInteger),
    Column('virtual', Boolean),
    schema="monitoring")

instances = Table(
    'instances', metadata,
    Column('instance_id', Integer, primary_key=True),
    Column('host_id', Integer,
           ForeignKey("monitoring.hosts.host_id"),
           nullable=False,
           ),
    Column('port', Integer, nullable=False),
    Column('local_name', UnicodeText, nullable=False,),
    Column('version', UnicodeText, nullable=False,),
    Column('version_num', Integer, nullable=False,),
    Column('data_directory', UnicodeText, nullable=False),
    Column('sysuser', UnicodeText,),
    Column('standby', Boolean, nullable=False, server_default=text('False')),
    UniqueConstraint('host_id', 'port'),
    schema="monitoring")

checks = Table(
    'checks', metadata,
    Column('check_id', Integer, primary_key=True),
    Column('host_id', Integer,
           ForeignKey("monitoring.hosts.host_id"),
           nullable=False,
           ),
    Column('instance_id', Integer,
           ForeignKey("monitoring.instances.instance_id"),
           nullable=True,
           ),
    Column('enabled', Boolean, nullable=False),
    Column('name', UnicodeText, nullable=False),
    Column('warning', REAL),
    Column('critical', REAL),
    Column('description', UnicodeText),
    schema="monitoring")

checkstates = Table(
    'check_states', metadata,
    Column('check_id', Integer,
           ForeignKey("monitoring.checks.check_id"),
           nullable=False, primary_key=True),
    Column('key', UnicodeText, nullable=True, primary_key=True),
    Column('state', UnicodeText, nullable=False),
    schema="monitoring")

collector_status = Table(
    'collector_status',
    metadata,
    Column('instance_id', Integer,
           ForeignKey("monitoring.instances.instance_id"),
           primary_key=True),
    Column('last_pull', DateTime, nullable=True),
    Column('last_push', DateTime, nullable=True),
    Column('last_insert', DateTime, nullable=True),
    Column('status', UnicodeText),
    schema="monitoring",
)



import csv
import operator


def get_csv_data(fd):
    # CSV parser as generator
    for r in csv.DictReader(fd):
        yield r


def pivot_timeserie(fd, index, key, value, output):
    # Simple pivot table implementation.
    # Beware, input data *MUST* be ordered by index value.
    fd.seek(0)
    keys = {}
    p = 1
    # We need to get the keys first
    for r in get_csv_data(fd):
        if r[key] not in keys:
            keys[r[key]] = p
            p += 1
    sk = sorted(keys.items(), key=operator.itemgetter(1))
    # CSV Header
    line = [index] + [x[0] for x in sk]
    p_index = ''
    fd.seek(0)
    for r in get_csv_data(fd):
        if r[index] != p_index:
            # As data are ordered if we meet a new index value then the current
            # line is complete.
            output.write(','.join(line)+'\n')
            # And start a new line
            line = [''] * (len(keys)+1)
            line[0] = r[index]
        # Append value to the current line
        line[keys[r[key]]] = r[value]
        # Keep a track of the index value
        p_index = r[index]
    # Write the last line
    output.write(','.join(line)+'\n')



import logging
import os.path
import sys
from time import sleep

import alembic.config
import sqlalchemy.exc
from alembic.runtime.environment import EnvironmentContext
from alembic.script import ScriptDirectory
from sqlalchemy.orm.session import sessionmaker
from sqlalchemy import create_engine

from ..toolkit.errors import UserError


Session = sessionmaker()
logger = logging.getLogger(__name__)


def format_dsn(dsn):
    fmt = "postgresql://{user}:{password}@:{port}/{dbname}?host={host}"
    return fmt.format(**dsn)


def build_alembic_config(temboard_config):
    config = alembic.config.Config()
    config.set_main_option(
        'sqlalchemy.url',
        format_dsn(temboard_config.repository),
    )
    config.set_main_option(
        'script_location',
        os.path.dirname(__file__) + '/alembic',
    )
    return config


def configure(dsn, **kwargs):
    if hasattr(dsn, 'items'):
        dsn = format_dsn(dsn)

    try:
        engine = create_engine(dsn)
        check_connectivity(engine)
    except Exception as e:
        logger.warning("Connection to the database failed: %s", e)
        logger.warning("Please check your configuration.")
        sys.stderr.write("FATAL: %s\n" % e)
        exit(10)
    Session.configure(bind=engine, **kwargs)

    # For legacy purpose, we return engine to ease binding other Session maker
    # with the same engine.
    return engine


def check_connectivity(engine):
    for i in range(10):
        try:
            engine.connect().close()
            break
        except Exception as e:
            if i == 9:
                raise
            logger.warn("Failed to connect to database: %s", e)
            logger.info("Retrying in %ss.", i)
            sleep(i)


def worker_engine(dbconf):
    """Create a new stand-alone SQLAlchemy engine to be instantiated in worker
    context.
    """
    return create_engine(format_dsn(dbconf))


def check_schema(config):
    # Derived from alembic.command.current.

    alembic_cfg = build_alembic_config(config)
    script = ScriptDirectory.from_config(alembic_cfg)

    def fn(current, context):
        target = script.get_current_head()
        current = context.get_current_revision()

        if current:
            logger.debug("temBoard database revision is %s.", current)
        else:
            logger.debug("temBoard database is uninitialized.")

        if current != target:
            raise UserError(
                "Database is not up to date. Please use temboard-migratedb."
            )
        else:
            logger.info("temBoard database is up-to-date.")

        return []  # Tells MigrationContext to skip migrations.

    with EnvironmentContext(alembic_cfg, script, fn=fn):
        try:
            script.run_env()
        except sqlalchemy.exc.OperationalError as e:
            raise UserError("Failed to check schema: %s." % e)



from temboardui.application import (
    send_mail,
    send_sms,
)
from temboardui.web import (
    HTTPError,
    admin_required,
    app,
    render_template,
)


@app.route(r'/settings/notifications')
@admin_required
def notifications(request):
    return render_template(
        'settings/notifications.html',
        nav=True, role=request.current_user,
        email_configured=app.config.notifications.smtp_host,
        sms_configured=app.config.notifications.twilio_account_sid,
    )


@app.route(r'/json/test_email', methods=['POST'])
@admin_required
def send_test_email(request):
    data = request.json
    email = data.get('email', None)
    if not email:
        raise HTTPError(400, "Email field is missing.")

    notifications_conf = app.config.notifications
    smtp_host = notifications_conf.smtp_host
    smtp_port = notifications_conf.smtp_port
    smtp_tls = notifications_conf.smtp_tls
    smtp_login = notifications_conf.smtp_login
    smtp_password = notifications_conf.smtp_password
    smtp_from_addr = notifications_conf.smtp_from_addr

    if not smtp_host:
        raise HTTPError(500, "SMTP server is not configured")

    send_mail(smtp_host, smtp_port, 'temBoard notification test',
              'This is a test', email, smtp_tls, smtp_login, smtp_password,
              smtp_from_addr)
    return {'message': 'OK'}


@app.route(r'/json/test_sms', methods=['POST'])
@admin_required
def send_test_sms(request):
    data = request.json
    phone = data.get('phone', None)
    if not phone:
        raise HTTPError(400, "Phone field is missing.")

    send_sms(app.config.notifications, 'temBoard notification test', [phone])
    return {'message': 'OK'}



"""statements

Revision ID: d2120f410bb5
Revises: 48dc1c585e8
Create Date: 2020-04-21 14:14:10.650409

"""
import os.path

from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = 'd2120f410bb5'
down_revision = '48dc1c585e8'
branch_labels = None
depends_on = None


def sqlfile(name):
    directory = os.path.dirname(__file__)
    with open(os.path.join(directory, name + '.sql')) as fo:
        return sa.text(fo.read())


def upgrade():
    op.get_bind().execute(sqlfile('statements'))



from ..web import (
    app,
    render_template,
)


@app.instance_route(r"/notifications")
def notifications(request):
    try:
        agent_username = request.instance.get_profile()['username']
    except Exception:
        agent_username = None
    notifications = request.instance.get("/notifications")
    return render_template(
        'notifications.html',
        agent_username=agent_username,
        instance=request.instance,
        nav=True,
        notifications=notifications,
        plugin='notifications',
        role=request.current_user,
    )



# coding: utf-8
from __future__ import absolute_import

import functools
import json
import logging
import os
import urllib2
from cStringIO import StringIO
from csv import writer as CSVWriter
from datetime import datetime

from tornado import web as tornadoweb
from tornado.concurrent import run_on_executor
from tornado.escape import json_decode, json_encode, url_escape
from tornado.gen import coroutine
from tornado.web import (
    Application as TornadoApplication,
    HTTPError,
    RequestHandler,
)
from tornado.template import Loader as TemplateLoader

from .application import (
    get_instance,
    get_role_by_cookie,
    get_roles_by_instance,
)
from .errors import TemboardUIError
from .model import Session as DBSession
from .temboardclient import (
    TemboardError,
    temboard_request,
)
from .toolkit.pycompat import PY2


logger = logging.getLogger(__name__)


def admin_required(func):
    # Similar to flask_security.roles_required, but limited to admin role.
    func.__admin_required = True
    return func


def anonymous_allowed(func):
    # Reverse of flask_security.login_required.
    #
    # In temboard, very few pages are anonymous. Thus we have implicit
    # login_required. This behaviour can be disabled by using
    # @anonymous_allowed.
    func.__anonymous_allowed = True
    return func


def serialize_querystring(query):
    return "&".join([
        "%s=%s" % (url_escape(name), url_escape(value))
        for name, value in sorted(query.items())
    ])


class Response(object):
    def __init__(
            self, status_code=200, headers=None, secure_cookies=None,
            body=u''):
        self.status_code = status_code
        self.headers = headers or {}
        self.secure_cookies = secure_cookies or {}
        self.body = body


class Redirect(Response, Exception):
    def __init__(self, location, permanent=False, secure_cookies=None):
        super(Redirect, self).__init__(
            status_code=301 if permanent else 302,
            headers={'Location': location},
            body=u'Redirected to %s' % location,
            secure_cookies=secure_cookies,
        )


class TemplateRenderer(object):
    # Flask-like HTML render function, without thread local.

    def __init__(self, path):
        self.loader = TemplateLoader(path)

    def __call__(self, template, **data):
        return Response(
            body=self.loader.load(template).generate(**data),
            headers={'Content-Type': 'text/html; charset=UTF-8'},
        )


template_path = os.path.realpath(__file__ + '/../templates')
render_template = TemplateRenderer(template_path)


def csvify(data, status_code=200):
    if isinstance(data, list):
        fo = StringIO()
        writer = CSVWriter(fo)
        for row in data:
            writer.writerow(row)
        data = fo.getvalue()
    elif not isinstance(data, (str, unicode)):
        raise ValueError("Malformed CSV data")
    return Response(
        status_code=status_code,
        headers={'Content-Type': 'text/csv'},
        body=data,
    )


def jsonify(data, status_code=200):
    return Response(
        status_code=status_code,
        headers={'Content-Type': 'application/json; charset=UTF-8'},
        body=json_encode(data),
    )


class CallableHandler(RequestHandler):
    # Adapt flask-like callable in Tornado Handler API.

    @property
    def executor(self):
        # To enable @run_on_executor methods, we must have executor property.
        return self.application.executor

    def initialize(self, callable_, blueprint=None, methods=None, logger=None):
        self.callable_ = callable_
        self.logger = logger or logging.getLogger(__name__)
        self.request.blueprint = blueprint
        self.request.config = self.application.config
        # run_on_executor searches for `executor` attribute of first argument.
        # Thus, we bind executor to request object.
        self.request.executor = self.executor
        self.request.handler = self
        self.SUPPORTED_METHODS = methods or ['GET']

    def get_current_user(self):
        cookie = self.get_secure_cookie('temboard')
        if not cookie:
            return

        try:
            return get_role_by_cookie(self.db_session, cookie)
        except Exception as e:
            self.logger.error("Failed to get role from cookie: %s ", e)

    @coroutine
    def get(self, *args, **kwargs):
        try:
            response = yield self.callable_(self.request, *args, **kwargs)
        except Redirect as response:
            pass

        if response is None:
            response = u''
        if isinstance(response, (dict, unicode)):
            response = Response(body=response)
        self.write_response(response)

    # Let's use one handler for all supported methods.
    post = get
    delete = get

    def write_response(self, response):
        # Should be in a middleware.
        if response.status_code in (301, 302, 401):
            response.secure_cookies['referer_uri'] = self.request.uri

        self.set_status(response.status_code)
        for k, v in response.headers.items():
            if not isinstance(v, list):
                v = [v]
            self.clear_header(k)
            for v1 in v:
                self.add_header(k, v1)

        for k, v in response.secure_cookies.items():
            self.set_secure_cookie(k, v, expires_days=30)

        self.finish(response.body)


class Error404Handler(RequestHandler):
    def write_error(self, status_code, **kwargs):
        content_type = self.request.headers.get("Content-Type", '')
        if content_type.startswith("application/json"):
            self.write({'error': 'Not found.'})
        else:
            self.write('404: Not found.')


class DatabaseHelper(object):
    @classmethod
    def add_middleware(cls, func):
        @functools.wraps(func)
        def database_middleware(request, *args):
            request.db_session = request.handler.db_session = DBSession()
            try:
                response = func(request, *args)
            except Exception:
                request.db_session.rollback()
                raise
            else:
                request.db_session.commit()
            finally:
                request.db_session.close()
                del request.db_session

            return response

        return database_middleware


class ErrorHelper(object):
    @classmethod
    def add_middleware(cls, func):
        @functools.wraps(func)
        def error_middleware(request, *args):
            try:
                return func(request, *args)
            except Redirect:
                raise
            except (TemboardError, TemboardUIError) as e:
                code = e.code
                message = e.message
            except HTTPError as e:
                code = e.status_code
                message = e.log_message
                if code == 404:
                    message += " Plugin may not be activated on agent side"
            except Exception as e:
                # Show traceback for developer, and HTML page for user.
                logger.exception("Unhandled Error:")
                code = 500
                message = str(e)

            logger.error("Request failed: %s %s.", code, message)
            return make_error(request, code, message)

        return error_middleware


class InstanceHelper(object):
    # This helper class implements all operations related to instance dedicated
    # request.

    INSTANCE_PARAMS = r'/(.*)/([0-9]{1,5})'
    PROXY_PREFIX = r'/proxy' + INSTANCE_PARAMS
    SERVER_PREFIX = r'/server' + INSTANCE_PARAMS

    @classmethod
    def add_middleware(cls, callable_):
        # Wraps an HTTP handler callable related to a Postgres instance

        @functools.wraps(callable_)
        def instance_middleware(request, address, port, *args):
            # Swallow adddress and port arguments.
            request.instance = cls(request)
            request.instance.fetch_instance(address, port)
            return callable_(request, *args)

        return instance_middleware

    def __init__(self, request):
        self.request = request
        self._xsession = False

    def __getattr__(self, name):
        return getattr(self.instance, name)

    def __repr__(self):
        return '<%s %s>' % (self.__class__.__name__, self.instance.hostname)

    def check_active_plugin(self, name):
        '''
        Ensure that the plugin is active for given instance
        '''
        if name not in [p.plugin_name for p in self.instance.plugins]:
            raise HTTPError(408, "Plugin %s not activated." % name)

    def fetch_instance(self, address, port):
        self.instance = get_instance(self.request.db_session, address, port)
        if not self.instance:
            raise HTTPError(404)

    @property
    def cookie_name(self):
        return 'temboard_%s_%s' % (
            self.instance.agent_address, self.instance.agent_port,
        )

    @property
    def xsession(self):
        if self._xsession is False:
            self._xsession = self.request.handler.get_secure_cookie(
                self.cookie_name)
        return self._xsession

    def format_url(self, path=''):
        return "/server/%s/%s%s" % (self.agent_address, self.agent_port, path)

    def redirect(self, path):
        raise Redirect(location=self.format_url(path))

    def http(self, path, method='GET', query=None, body=None):
        url = 'https://%s:%s%s?key=%s' % (
            self.instance.agent_address,
            self.instance.agent_port,
            path,
            self.instance.agent_key,
        )

        if query:
            url += "&" + serialize_querystring(query)

        headers = {}
        xsession = self.xsession
        if xsession:
            headers['X-Session'] = xsession

        logger.debug("Proxying %s %s.", method, url)
        try:
            body = temboard_request(
                self.request.config.temboard.ssl_ca_cert_file,
                method=method,
                url=url,
                headers=headers,
                data=body,
            )
        except urllib2.HTTPError as e:
            message = e.read()
            try:
                message = json_decode(message)['error']
            except Exception as ee:
                logger.debug("Failed to decode agent error: %s.", ee)
            raise HTTPError(e.code, message)
        except urllib2.URLError as e:
            logger.error("Proxied request failed: %s", e)
            raise HTTPError(500, str(e.reason))
        except Exception as e:
            logger.error("Proxied request failed: %s", e)
            raise HTTPError(500)
        return json_decode(body)

    def get(self, *args, **kwargs):
        kwargs['method'] = 'GET'
        return self.http(*args, **kwargs)

    def post(self, *args, **kwargs):
        kwargs['method'] = 'POST'
        return self.http(*args, **kwargs)

    def require_xsession(self):
        if not self.xsession:
            self.redirect('/login')
        return self.xsession

    def get_profile(self):
        try:
            self.require_xsession()
            return self.get("/profile")
        except HTTPError as e:
            if 401 == e.status_code:
                self.redirect('/login')
            raise


def add_json_middleware(func):
    @functools.wraps(func)
    def json_middleware(request, *args):
        if 'POST' == request.method:
            try:
                request.json = json_decode(request.body)
            except Exception as e:
                raise HTTPError(400, e.message)
        else:
            # Set empty body, make_error catch this to format error as JSON.
            request.json = None

        return func(request, *args)
    return json_middleware


def add_user_instance_middleware(func):
    # Ensures user is allowed to access to the instance
    @functools.wraps(func)
    def user_instance_middleware(request, address, port, *args):
        user = request.current_user = request.handler.current_user

        if user is None:
            # Not logged in
            raise HTTPError(401, "Restricted area.")

        allowed_roles = get_roles_by_instance(request.handler.db_session,
                                              address,
                                              port)
        roles = [role.role_name for role in allowed_roles if role]
        if user.role_name not in roles:
            raise HTTPError(403, "Restricted area.")

        return func(request, address, port, *args)

    return user_instance_middleware


class UserHelper(object):
    @classmethod
    def add_middleware(cls, func):

        @functools.wraps(func)
        def user_middleware(request, *args):
            role = request.current_user = request.handler.current_user

            anonymous_allowed = getattr(func, '__anonymous_allowed', False)
            if not anonymous_allowed and role is None:
                logger.debug("Redirecting anonymous to /login.")
                raise Redirect('/login')

            admin_required = getattr(func, '__admin_required', False)
            if admin_required and not role.is_admin:
                logger.debug("Refusing access to non-admin user.")
                raise HTTPError(403)

            return func(request, *args)

        return user_middleware


# Ensure @functools.wraps preserves User middleware attributes.
functools.WRAPPER_UPDATES += (
    '__admin_required',
    '__anonymous_allowed',
)


class Blueprint(object):
    def __init__(self, plugin_name=None):
        self.plugin_name = plugin_name
        self.rules = []

    def add_rules(self, rules):
        self.rules.extend(rules)

    def generic_proxy(self, url, methods=None):
        # Pass-through implementation for /proxy/address/port/…
        url = r'(%s)' % url

        @self.instance_proxy(url, methods)
        def generic_instance_proxy(request, path):
            if request.blueprint and request.blueprint.plugin_name:
                request.instance.check_active_plugin(
                    request.blueprint.plugin_name)
            body = request.instance.http(
                path=url_escape(path, plus=False),
                method=request.method,
                body=request.json,
            )
            return jsonify(body)

    def instance_proxy(self, url, methods=None):
        # decorator for /proxy/address/port/… handlers.
        url = InstanceHelper.PROXY_PREFIX + url
        return self.route(url, methods=methods, with_instance=True, json=True)

    def instance_route(self, url, methods=None, **kwargs):
        # Helper to declare a route with instance URL prefix and middleware.
        return self.route(
            url=InstanceHelper.SERVER_PREFIX + url,
            methods=methods,
            with_instance=True,
            **kwargs
        )

    def route(self, url, methods=None, with_instance=False, json=None):
        # Implements flask-like route registration of a simple synchronous
        # callable.

        # Enable JSON middleware on /json/ handlers.
        if json is None:
            json = url.startswith('/json/')

        def decorator(func):
            logger_name = func.__module__ + '.' + func.__name__

            func = UserHelper.add_middleware(func)
            if with_instance:
                func = InstanceHelper.add_middleware(func)

                if url.startswith('/server/') or url.startswith('/proxy/'):
                    # Limit user/instance access control to /server/ and
                    # /proxy/.
                    # Admin area access control has already been performed
                    func = add_user_instance_middleware(func)

            if json:
                func = add_json_middleware(func)
            func = ErrorHelper.add_middleware(func)
            func = DatabaseHelper.add_middleware(func)

            @run_on_executor
            @functools.wraps(func)
            def sync_request_wrapper(request, *args):
                try:
                    return func(request, *args)
                except (Redirect, HTTPError):
                    raise
                except Exception as e:
                    # Since async traceback is useless, spit here traceback and
                    # forge simple HTTPError(500), no HTML error.
                    logger.exception("Unhandled Error:")
                    raise HTTPError(500, str(e))

            rules = [(
                url, CallableHandler, dict(
                    blueprint=self,
                    callable_=sync_request_wrapper,
                    methods=methods or ['GET'],
                    logger=logging.getLogger(logger_name),
                ),
            )]
            self.add_rules(rules)
            return func

        return decorator


class WebApplication(TornadoApplication, Blueprint):
    def __init__(self, *a, **kwargs):
        super(WebApplication, self).__init__(*a, **kwargs)

    def configure(self, **settings):
        # Runtime configuration of application.
        #
        # This way, we can initialize app at import time to register handlers.
        # Then configure it at run time once configuration is parsed.

        self.settings.update(settings)

        # This comes from Tornado's __init__
        if self.settings.get('debug'):
            self.settings.setdefault('autoreload', True)
            self.settings.setdefault('compiled_template_cache', False)
            self.settings.setdefault('static_hash_cache', False)
            self.settings.setdefault('serve_traceback', True)

    def add_rules(self, rules):
        if hasattr(self, 'wildcard_router'):  # Tornado 4.5+
            self.wildcard_router.add_rules(rules)
        elif not self.handlers:
            self.add_handlers(r'.*$', rules)
        else:
            rules = [tornadoweb.URLSpec(*r) for r in rules]
            self.handlers[0][1].extend(rules)


def make_error(request, code, message):
    # If ?noerror=1 is set, only return HTTP error code.
    if "1" == request.handler.get_argument('noerror', None):
        logger.debug("Hide error %s %s for ?noerror=1.", code, message)
        return Response(200)

    data = dict(code=code, error=message)

    if hasattr(request, 'json'):
        return Response(code, body=data)

    if hasattr(request, 'instance'):
        data['instance'] = request.instance

    template = 'unauthorized.html' if 403 == code else 'error.html'
    response = render_template(
        template,
        nav=True, role=request.current_user,
        **data
    )
    response.status_code = code
    return response


class JSONEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, datetime):
            return obj.isoformat()
        else:
            return super(JSONEncoder, self).default(obj)


# Change default cls argument to custom encoder.
if PY2:
    defaults = list(json.dumps.func_defaults)
    defaults[4] = JSONEncoder
    json.dumps.func_defaults = tuple(defaults)
else:
    json.dumps.__kwdefaults__['cls'] = JSONEncoder


# Global app instance for registration of core handlers.
app = WebApplication()
# Hijack tornado.web access_log to log request in temboardui namespace.
tornadoweb.access_log = logging.getLogger('temboardui.access')



"""alerting

Revision ID: cd5b518ab5cf
Revises: 81f0aef8a119
Create Date: 2020-03-25 16:56:26.929549

"""
import os.path

from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = 'cd5b518ab5cf'
down_revision = '81f0aef8a119'
branch_labels = None
depends_on = None


def sqlfile(name):
    directory = os.path.dirname(__file__)
    with open(os.path.join(directory, name + '.sql')) as fo:
        return sa.text(fo.read())


def upgrade():
    op.get_bind().execute(sqlfile('alerting'))



# -*- coding: utf-8 -*-
#
# This module implements protocol switching from HTTP to HTTPS on the same
# socket connection.
#
# In tornado, a bunch of objects interacts with each other to handle a request:
# IOStream, TCP/HTTPServer, HTTP1Connection, HTTPMessageDelegate, HTTPRequest,
# HTTPResponse and HTTPRequestHandler. But it's not as simple as an onion of
# successive encapsulations.
#
# OpenSSL raises an error on uncrypted HTTP request while waiting for
# handshake. The first step is to have an IOStream with explicit SSL handshake
# allowing to catch this OpenSSL error. This stream must NOT close the socket
# on error.
#
# On HTTP request, we divert the request processing to the protocol switcher.
# The protocol switcher almost look like a regular request/response handling
# but there is no routing. We limit processing to headers to preserver host and
# path of the request.
#
# All of this (handshake, read request and write response) must be async to
# keep the server loop managing all incoming connections.

import errno
import logging
import socket
import ssl

from tornado import gen
from tornado.httpserver import HTTPServer, HTTPRequest
from tornado.httpclient import HTTPResponse
from tornado.httputil import (
    HTTPHeaders,
    parse_request_start_line,
    ResponseStartLine,
)
from tornado.http1connection import HTTP1Connection
from tornado.ioloop import IOLoop
from tornado.iostream import (
    IOStream,
    SSLIOStream,
    StreamClosedError,
)
from tornado.log import (
    app_log,
    gen_log,
)
from tornado.netutil import ssl_wrap_socket
from tornado.util import errno_from_exception


logger = logging.getLogger(__name__)


def parse_http_headers(payload):
    # Implements simple HTTP1Connection._read_message but IO-free.
    lines = payload.splitlines()
    if lines and ':' not in lines[0]:
        # Drop start line
        lines.pop(0)
    # Drop contents
    if '' in lines:
        lines[:] = lines[:lines.index('')]
    return (
        parse_request_start_line('GET / HTTP/1.1'),
        HTTPHeaders.parse('\r\n'.join(lines)),
    )


def protocol_switcher(request):
    try:
        host = request.headers['Host']
    except KeyError:
        # We don't have FQDN. Fallback to socket address. This breaks
        # name-based virtualhost.
        host = '%(address)s:%(port)s' % dict(
            request.config.temboard, address=request.host)
    new_url = 'https://%s%s' % (host, request.uri)
    headers = HTTPHeaders({
        'Content-Length': '0',
        'Location': new_url,
    })
    logger.debug("Redirecting client to %s.", new_url)
    return HTTPResponse(
        request=request, code=301, headers=headers,
        # If effective_url is not set, HTTPResponse falls back to request.url,
        # which does not exists... See tornado.httpclient.HTTPResponse.__init__
        # and tornado.httpserver.HTTPRequest.
        effective_url=request.full_url(),
    )


class SSLErrorHTTPRequest(Exception):
    pass


class EasySSLIOStream(SSLIOStream):
    # SSIOStream raising exception on HTTP_REQUEST rather than closing socket.

    def _do_ssl_handshake(self):
        # Based on code from test_ssl.py in the python stdlib
        try:
            self._handshake_reading = False
            self._handshake_writing = False
            self.socket.do_handshake()
        except ssl.SSLError as err:
            if err.args[0] == ssl.SSL_ERROR_WANT_READ:
                self._handshake_reading = True
                return
            elif err.args[0] == ssl.SSL_ERROR_WANT_WRITE:
                self._handshake_writing = True
                return
            elif err.args[0] in (ssl.SSL_ERROR_EOF,
                                 ssl.SSL_ERROR_ZERO_RETURN):
                return self.close(exc_info=True)
            elif err.args[0] == ssl.SSL_ERROR_SSL:
                try:
                    peer = self.socket.getpeername()
                except Exception:
                    peer = '(not connected)'

                if getattr(err, 'reason', None) == 'HTTP_REQUEST':
                    # Async raise HTTP_REQUEST error.
                    self._ssl_connect_future.set_exception(
                        SSLErrorHTTPRequest()
                    )
                    gen_log.warning("HTTP_REQUESTS on SSL handshake from %s.",
                                    peer)
                    return

                gen_log.warning("SSL Error on %s %s: %s",
                                self.socket.fileno(), peer, err)
                return self.close(exc_info=True)
            raise
        except socket.error as err:
            # Some port scans (e.g. nmap in -sT mode) have been known
            # to cause do_handshake to raise EBADF and ENOTCONN, so make
            # those errors quiet as well.
            # https://groups.google.com/forum/?fromgroups#!topic/python-tornado/ApucKJat1_0
            if (self._is_connreset(err) or
                    err.args[0] in (errno.EBADF, errno.ENOTCONN)):
                return self.close(exc_info=True)
            raise
        except AttributeError:
            # On Linux, if the connection was reset before the call to
            # wrap_socket, do_handshake will fail with an
            # AttributeError.
            return self.close(exc_info=True)
        else:
            self._ssl_accepting = False
            if not self._verify_cert(self.socket.getpeercert()):
                self.close()
                return
            self._run_ssl_connect_callback()


class AutoHTTPSServer(HTTPServer):
    # HTTPServer implementing protocol switching.

    def _handle_connection(self, connection, address):
        # Copy-paste of tornado.httpserver.HTTPServer._handle_connection to use
        # our custom stream class.
        #
        # Actually, connection is just a socket.
        try:
            connection = ssl_wrap_socket(connection,
                                         self.ssl_options,
                                         server_side=True,
                                         do_handshake_on_connect=False)
        except ssl.SSLError as err:
            if err.args[0] == ssl.SSL_ERROR_EOF:
                return connection.close()
            else:
                raise
        except socket.error as err:
            # If the connection is closed immediately after it is created
            # (as in a port scan), we can get one of several errors.
            # wrap_socket makes an internal call to getpeername,
            # which may return either EINVAL (Mac OS X) or ENOTCONN
            # (Linux).  If it returns ENOTCONN, this error is
            # silently swallowed by the ssl module, so we need to
            # catch another error later on (AttributeError in
            # SSLIOStream._do_ssl_handshake).
            # To test this behavior, try nmap with the -sT flag.
            # https://github.com/tornadoweb/tornado/pull/750
            if errno_from_exception(err) in (errno.ECONNABORTED, errno.EINVAL):  # noqa
                return connection.close()
            else:
                raise
        try:
            io_loop = self.io_loop
            kw = dict(io_loop=io_loop)
        except AttributeError:
            # We are on Tornado 5+. Just don't pass ioloop
            kw = {}
            io_loop = IOLoop.current()

        try:
            stream = EasySSLIOStream(
                connection,
                max_buffer_size=self.max_buffer_size,
                read_chunk_size=self.read_chunk_size,
                **kw
            )
            future = self.handle_stream(stream, address)
            if future is not None:
                io_loop.add_future(future, lambda f: f.result())
        except Exception:
            app_log.error("Error in connection callback", exc_info=True)

    @gen.coroutine
    def handle_stream(self, ssl_stream, address):
        try:
            yield ssl_stream.wait_for_handshake()
        except SSLErrorHTTPRequest:
            stream = IOStream(ssl_stream.socket._sock)
            conn = HTTP1Connection(stream, is_client=False)
            try:
                yield self.handle_http_connection(conn)
            except Exception:
                logger.exception("Failed to process HTTP request:")
            finally:
                stream.close()
        except StreamClosedError:
            logger.debug("Stream closed by client during handshake. Skipping.")
            return
        else:
            super(AutoHTTPSServer, self).handle_stream(ssl_stream, address)

    @gen.coroutine
    def handle_http_connection(self, conn):
        # Read the trailing HTTP request and process it with protocol_switcher.
        # We can't rely on ioloop to trigger read because it has been already
        # triggered for SSL handshake.
        addr, port = conn.stream.socket.getsockname()
        try:
            # This is not blocking. Just read available bytes.
            payload = conn.stream.socket.recv(1024)
        except Exception:
            # Exception includes EWOULDBLOCK, when no bytes are available. In
            # this case just skip.
            payload = ""
        else:
            logger.debug("Received %r", payload[:128])
        # Simulate conn._read_message side effect. This is required by
        # HTTP1Connection.write_headers()
        conn._request_start_line = parse_request_start_line('GET / HTTP/1.1')
        try:
            start_line, headers = parse_http_headers(payload)
            conn._request_start_line = start_line
            request = HTTPRequest(
                connection=conn,
                headers=headers,
                start_line=start_line,
            )
            request.config = self.request_callback.config
            response = protocol_switcher(request)
        except Exception as e:
            logger.error("Failed to switch to HTTPS: %s", e)
            response = HTTPResponse(
                request=object(), code=500,
                headers=HTTPHeaders({'Content-Length': '0'}),
                effective_url='https://useless_effective_url'
            )
        yield conn.write_headers(
            start_line=ResponseStartLine(
                'HTTP/1.1', response.code, response.reason,
            ),
            headers=response.headers,
        )



import datetime
from sqlalchemy import (
    Boolean,
    DateTime,
    Integer,
    String,
    event,
)
from sqlalchemy.orm import relationship
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.sql import (
    case,
    column,
    extract,
    func,
)

from temboardui.model import tables

Model = declarative_base()


def validate_int(value):
    if value is None:
        return
    if isinstance(value, unicode):
        value = int(value)
    if isinstance(value, basestring):
        value = int(value)
    assert isinstance(value, int)
    return value


def validate_string(value):
    assert isinstance(value, basestring)
    return value


def validate_datetime(value):
    assert isinstance(value, datetime.datetime)
    return value


def validate_boolean(value):
    if isinstance(value, int):
        if value == 1:
            value = True
        if value == 0:
            value = False
    if isinstance(value, basestring):
        if str(value).lower() == 'yes' or \
           str(value).lower() == 'true' or \
           str(value).lower() == 't' or value == '1':
            value = True
        if str(value).lower() == 'no' or \
           str(value).lower() == 'false' or \
           str(value).lower() == 'f' or value == '0':
            value = False
    assert isinstance(value, bool)
    return value


validators = {
    Integer: validate_int,
    String: validate_string,
    DateTime: validate_datetime,
    Boolean: validate_boolean
}


def diff(var):
    return (func.max(column(var)) - func.min(column(var))).label(var)


def to_epoch(column):
    return extract("epoch", column).label(column.name)


def total_measure_interval(column):
    return extract(
        "epoch",
        case([(func.min(column) == '0 second', '1 second')],
             else_=func.min(column)))


# We use 8192 as default value for block size
# Ideally we should get this value from agent
block_size = 8192


def total_read(c):
    return (
        func.sum(c.shared_blks_read + c.local_blks_read + c.temp_blks_read) /
        total_measure_interval(c.mesure_interval)
    ).label("total_blks_read")


def total_hit(c):
    return (
        func.sum(c.shared_blks_hit + c.local_blks_hit) /
        total_measure_interval(c.mesure_interval)
    ).label("total_blks_hit")


@event.listens_for(Model, 'attribute_instrument')
def configure_listener(class_, key, inst):
    if not hasattr(inst.property, 'columns'):
        return

    @event.listens_for(inst, "set", retval=True)
    def set_(instance, value, oldvalue, initiator):
        validator = validators.get(inst.property.columns[0].type.__class__)
        if validator:
            try:
                return validator(value)
            except AssertionError:
                raise Exception("%s: wrong type." % (inst.property))
        else:
            return value


class Plugins(Model):
    __table__ = tables.plugins


class InstanceGroups(Model):
    __table__ = tables.instance_groups


class RoleGroups(Model):
    __table__ = tables.role_groups


class AccessRoleInstance(Model):
    __table__ = tables.access_role_instance


class Roles(Model):
    __table__ = tables.roles
    groups = relationship(
        RoleGroups,
        order_by=RoleGroups.group_name,
        backref="roles",
        cascade="save-update, merge, delete, delete-orphan"
    )


class Instances(Model):
    __table__ = tables.instances
    groups = relationship(
        InstanceGroups,
        order_by="InstanceGroups.group_name",
        backref="instances",
        cascade="save-update, merge, delete, delete-orphan"
    )

    plugins = relationship(
        Plugins,
        order_by="Plugins.plugin_name",
        backref="instances",
        cascade="save-update, merge, delete, delete-orphan"
    )


class Groups(Model):
    __table__ = tables.groups
    ari = relationship(
        AccessRoleInstance,
        order_by=AccessRoleInstance.role_group_name,
        backref="groups",
        cascade="save-update, merge, delete, delete-orphan",
        foreign_keys=[
            AccessRoleInstance.instance_group_name,
            AccessRoleInstance.instance_group_kind
        ]
    )


class Biggest(object):

    def __init__(self, order_by):
        self.order_by = order_by

    def __call__(self, var, minval=0, label=None):
        label = label or var
        return func.greatest(
            column(var) -
            func.lag(column(var))
            .over(order_by=self.order_by),
            minval
        ).label(label)


class Biggestsum(object):

    def __init__(self, order_by):
        self.order_by = order_by

    def __call__(self, var, minval=0, label=None):
        label = label or var
        return func.greatest(
            func.sum(column(var)) -
            func.lag(func.sum(column(var)))
            .over(order_by=self.order_by),
            minval
        ).label(label)



from os import path
import tornado.web

from temboardui.web import (
    Blueprint,
    TemplateRenderer,
)


PLUGIN_NAME = 'activity'
blueprint = Blueprint()
blueprint.generic_proxy(r'/activity/kill', methods=['POST'])
plugin_path = path.dirname(path.realpath(__file__))
render_template = TemplateRenderer(plugin_path + '/templates')


def configuration(config):
    return {}


def get_routes(config):
    routes = blueprint.rules + [
        (r"/js/activity/(.*)", tornado.web.StaticFileHandler, {
            'path': plugin_path + "/static/js"
        }),
    ]
    return routes


def get_agent_username(request):
    try:
        return request.instance.get_profile()['username']
    except Exception:
        return None


@blueprint.instance_route(r'/activity/(running|blocking|waiting)')
def activity(request, mode):
    request.instance.check_active_plugin(PLUGIN_NAME)
    agent_username = get_agent_username(request)
    xsession = request.instance.xsession if agent_username else None
    return render_template(
        'activity.html',
        nav=True,
        agent_username=agent_username,
        instance=request.instance,
        plugin=PLUGIN_NAME,
        mode=mode,
        xsession=xsession,
        role=request.current_user,
    )


@blueprint.instance_proxy(r'/activity(?:/blocking|/waiting)?')
def activity_proxy(request):
    request.instance.check_active_plugin(PLUGIN_NAME)
    return dict(
        blocking=request.instance.get('/activity/blocking'),
        running=request.instance.get('/activity'),
        waiting=request.instance.get('/activity/waiting'),
    )



"""Add comment field to instances table

Revision ID: 48dc1c585e8
Revises: cd5b518ab5cf
Create Date: 2020-07-06 11:02:05.598278

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '48dc1c585e8'
down_revision = 'cd5b518ab5cf'
branch_labels = None
depends_on = None


def upgrade():
    op.add_column('instances',
                  sa.Column('comment', sa.UnicodeText),
                  schema='application')


def downgrade():
    op.drop_column('instances', 'comment', schema='application')



import logging
import sys

logger = logging.getLogger(__name__)


def load_plugins(plugin_names, config):
    """
    Intend to load plugins and run their configuration() function.
    Plugins are defined as a module located in plugins/ directory. Plugins list
    to load is declared into temboard section of the configuration file:
        [temboard]
        plugins = [ "plugin1", "plugin2" ]
    """

    ret = dict()
    # Loop through declared plugins.
    for plugin_name in plugin_names:
        try:
            __import__('temboardui.plugins', fromlist=[plugin_name])
            module = sys.modules['temboardui.plugins.' + plugin_name]
        except Exception:
            logger.exception("Failed to load %s module.", plugin_name)
            continue

        # Try to run module's configuration() function.
        try:
            ret.update({plugin_name: {
                'configuration': getattr(module, 'configuration')(config),
                'routes': getattr(module, 'get_routes')(config),
                'workers': getattr(module, 'workers', None),
            }})
        except Exception:
            logger.exception("Failed to load %s configuration.", plugin_name)
            continue
        logger.info("Loaded plugin '%s'." % (plugin_name, ))

    return ret



import logging
from time import sleep

import tornado.web

from ..application import (
    gen_cookie,
    get_role_by_auth,
    hash_password,
)
from ..errors import TemboardUIError
from ..web import (
    HTTPError,
    Redirect,
    Response,
    anonymous_allowed,
    app,
    render_template,
)


logger = logging.getLogger(__name__)


@app.route('/logout')
def logout(request):
    request.handler.clear_cookie('temboard')
    # Redirect to /home so that /login referer is /home, not logout.
    return Redirect('/home')


def login_common(db_session, rolename, password):
    # Common logic between json and HTML login.
    passhash = hash_password(rolename, password)
    role = get_role_by_auth(db_session, rolename, passhash)
    logger.info(u"Role '%s' authentificated.", role.role_name)
    return dict(temboard=gen_cookie(role.role_name, passhash))


@app.route('/login', methods=['GET', 'POST'])
@anonymous_allowed
def login(request):
    if 'GET' == request.method:
        if request.current_user is None:
            return render_template('login.html', nav=False)
        else:
            return Redirect('/home')
    else:
        # Ensure request take at least one second to mitigate dictionnaries
        # attacks.
        sleep(1)
        try:
            rolename = request.handler.get_argument('username')
            password = request.handler.get_argument('password')
            cookies = login_common(request.db_session, rolename, password)
            referer = request.handler.get_secure_cookie('referer_uri')
            return Redirect(referer or '/home', secure_cookies=cookies)
        except TemboardUIError as e:
            logger.error(u"Login failed: %s", e)
            response = render_template(
                'login.html',
                nav=False,
                error=u"Wrong username/password.",
            )
            response.status_code = 401
            return response


@app.route(r'/json/login', methods=['POST'])
@anonymous_allowed
def json_login(request):
    # Mitigate dictionnaries attacks.
    sleep(1)
    try:
        post = tornado.escape.json_decode(request.body)
        username = post['username']
        password = post['password']
        cookies = login_common(request.db_session, username, password)
        return Response(
            body={"message": "OK"},
            secure_cookies=cookies,
        )
    except TemboardUIError as e:
        logger.error(u"Login failed: %s", e)
        return Response(
            status_code=401,
            body={"error": "Wrong username/password."},
        )


@app.instance_route(r'/login', methods=['GET', 'POST'])
def agent_login(request):
    error = None
    username = None
    if 'GET' == request.method:
        if not request.current_user:
            return Redirect('/login')

        redirect_to = request.handler.get_query_argument('redirect_to', None)
        if redirect_to:
            request.handler.set_secure_cookie('referer_uri', redirect_to,
                                              expires_days=30)
        if request.instance.xsession:
            try:
                profile = request.instance.get_profile()
            except Exception as e:
                logger.debug("Failed to get profile: %s.", e)
                logger.debug("Is X-Session valid? Showing login form.")
            else:
                username = profile['username']
    else:
        creds = dict(
            username=request.handler.get_argument('username'),
            password=request.handler.get_argument('password'),
        )
        try:
            response = request.instance.post("/login", body=creds)
            xsession = response['session']
        except HTTPError as e:
            error = e.log_message
        else:
            cookies = {request.instance.cookie_name: xsession}
            location = request.handler.get_secure_cookie('referer_uri')
            if not location:
                location = request.instance.format_url('/dashboard')
            return Redirect(location, secure_cookies=cookies)

    return render_template(
        'agent-login.html',
        nav=True, instance=request.instance, username=username,
        role=request.current_user,
        error=error,
    )



from binascii import hexlify
from hashlib import sha512
import base64
import json
import logging
import re
import urllib
import urllib2

from sqlalchemy.orm import joinedload
from sqlalchemy.orm.exc import (
    NoResultFound,
)
from sqlalchemy.exc import (
    IntegrityError,
)
from smtplib import SMTP, SMTP_SSL
from email.mime.text import MIMEText

from temboardui.model.orm import (
    AccessRoleInstance,
    Groups,
    Instances,
    InstanceGroups,
    Plugins,
    Roles,
    RoleGroups,
)
from temboardui.errors import TemboardUIError

logger = logging.getLogger(__name__)

"""
Roles
"""


def add_role(session,
             role_name,
             role_password,
             role_email,
             is_active=True,
             is_admin=False):
    try:
        role = Roles(
            role_name=unicode(role_name),
            role_password=unicode(role_password),
            role_email=unicode(role_email),
            is_active=is_active,
            is_admin=is_admin)
        session.add(role)
        session.flush()
        return role
    except IntegrityError as e:
        if e.message.find('roles_role_email_key') > 0:
            raise TemboardUIError(400, "Email address '%s' already in use." %
                                  (role_email))
        elif e.message.find('roles_pkey') > 0:
            raise TemboardUIError(400,
                                  "Role '%s' already exists." % (role_name))
        else:
            raise TemboardUIError(e.message)
    except Exception as e:
        raise TemboardUIError(e.message)


def update_role(session,
                role_name,
                new_role_name=None,
                role_password=None,
                role_email=None,
                is_active=None,
                is_admin=None,
                role_phone=None):
    try:
        role = session.query(Roles) \
            .filter_by(role_name=unicode(role_name)) \
            .first()
        if new_role_name is not None:
            role.role_name = unicode(new_role_name)
        if role_password is not None:
            role.role_password = unicode(role_password)
        if role_email is not None:
            role.role_email = unicode(role_email)
        if role_phone is not None:
            role.role_phone = unicode(role_phone) if role_phone else None
        if is_active is not None:
            role.is_active = is_active
        if is_admin is not None:
            role.is_admin = is_admin
        session.merge(role)
        session.flush()
        return role
    except IntegrityError as e:
        if e.message.find('roles_role_email_key') > 0:
            raise TemboardUIError(400, "Email address '%s' already in use." %
                                  (role_email))
        elif e.message.find('roles_pkey') > 0:
            raise TemboardUIError(400, "Role '%s' already exists." %
                                  (new_role_name))
        else:
            raise TemboardUIError(e.message)
    except AttributeError as e:
        raise TemboardUIError(400, "Role '%s' not found." % (role_name))
    except Exception as e:
        raise TemboardUIError(500, e.message)


def get_role(session, role_name):
    try:
        return session.query(Roles).options(
            joinedload(Roles.groups)).filter_by(
                role_name=unicode(role_name)).first()
    except AttributeError as e:
        raise TemboardUIError(400, "Role '%s' not found." % (role_name))
    except Exception as e:
        raise TemboardUIError(e.message)


def delete_role(session, role_name):
    try:
        role = session.query(Roles).filter(
            Roles.role_name == unicode(role_name)).one()
        session.delete(role)
    except NoResultFound as e:
        raise TemboardUIError(400, "Role '%s' not found." % (role_name))
    except Exception as e:
        raise TemboardUIError(e.message)


def get_role_list(
        session, ):
    return session.query(Roles).options(joinedload(Roles.groups)).order_by(
        Roles.role_name).all()


"""
Groups
"""


def add_group(session, group_name, group_description, group_kind):
    try:
        group = Groups(
            group_name=unicode(group_name),
            group_description=unicode(group_description),
            group_kind=unicode(group_kind))
        session.add(group)
        session.flush()
        return group
    except IntegrityError as e:
        if e.message.find('groups_group_kind_check') > 0:
            raise TemboardUIError(400, "Group kind '%s' does not exist." %
                                  (group_kind))
        elif e.message.find('groups_pkey') > 0:
            raise TemboardUIError(400, "Group '%s' ('%s') already exists." %
                                  (group_name, group_kind))
        else:
            raise TemboardUIError(400, e.message)
    except Exception as e:
        raise TemboardUIError(400, e.message)


def get_group(session, group_name, group_kind):
    try:
        if group_kind == 'role':
            return session.query(Groups).filter_by(
                group_name=unicode(group_name),
                group_kind=unicode(group_kind)).one()
        else:
            return session.query(Groups).options(
                joinedload(Groups.ari)).filter(
                    Groups.group_name == unicode(group_name),
                    Groups.group_kind == unicode(group_kind)).one()
    except AttributeError as e:
        raise TemboardUIError(400, "Group '%s' (%s) not found." % (group_name,
                                                                   group_kind))
    except Exception as e:
        raise TemboardUIError(e.message)


def update_group(session,
                 group_name,
                 group_kind,
                 new_group_name=None,
                 group_description=None):
    try:
        group = session.query(Groups) \
            .filter_by(
                group_name=unicode(group_name),
                group_kind=unicode(group_kind)) \
            .first()
        if new_group_name is not None:
            group.group_name = unicode(new_group_name)
        if group_description is not None:
            group.group_description = unicode(group_description)
        session.merge(group)
        session.flush()
        return group
    except IntegrityError as e:
        if e.message.find('groups_pkey') > 0:
            raise TemboardUIError(400, "Group name '%s' already in use." %
                                  (new_group_name))
        else:
            raise TemboardUIError(400, e.message)
    except AttributeError as e:
        raise TemboardUIError(400, "Group '%s' ('%s') not found." %
                              (group_name, group_kind))
    except Exception as e:
        raise TemboardUIError(400, e.message)


def delete_group(session, group_name, group_kind):
    try:
        group = session.query(Groups).filter(
            Groups.group_name == unicode(group_name),
            Groups.group_kind == unicode(group_kind)).one()
        session.delete(group)
    except NoResultFound as e:
        raise TemboardUIError(400, "Group '%s' not found." % (group_name))
    except Exception as e:
        raise TemboardUIError(400, e.message)


def get_group_list(session, group_kind='role'):
    if group_kind == 'role':
        return session.query(Groups).filter(
            Groups.group_kind == unicode(group_kind)).order_by(
                Groups.group_name).all()
    else:
        return session.query(Groups).options(joinedload(Groups.ari)).filter(
            Groups.group_kind == unicode(group_kind)).order_by(
                Groups.group_name).all()


def add_role_in_group(session, role_name, group_name):
    try:
        role_group = RoleGroups(
            role_name=unicode(role_name), group_name=unicode(group_name))
        session.add(role_group)
        session.flush()
    except IntegrityError as e:
        if e.message.find('role_groups_group_name_fkey') > 0:
            raise TemboardUIError(400, "Group '%s' ('role') does not exist." %
                                  (group_name))
        elif e.message.find('role_groups_role_name_fkey') > 0:
            raise TemboardUIError(400,
                                  "Role '%s' does not exist." % (role_name))
        elif e.message.find('role_groups_pkey') > 0:
            raise TemboardUIError(400, "Role '%s' already in group '%s'." %
                                  (role_name, group_name))
        else:
            raise TemboardUIError(400, e.message)
    except Exception as e:
        raise TemboardUIError(400, e.message)


def delete_role_from_group(session, role_name, group_name):
    try:
        role_group = session.query(RoleGroups).filter(
            RoleGroups.group_name == unicode(group_name),
            RoleGroups.role_name == unicode(role_name)).one()
        session.delete(role_group)
    except NoResultFound as e:
        raise TemboardUIError(400, "Role '%s' not found in group '%s'." %
                              (role_name, group_name))
    except Exception as e:
        raise TemboardUIError(400, e.message)


def get_groups_by_role(session, role_name):
    return session.query(RoleGroups).filter(
        RoleGroups.role_name == unicode(role_name)).order_by(
            RoleGroups.group_name).all()


def get_instance_groups_by_role(session, role_name):
    return session.query(InstanceGroups.group_name).filter(
        InstanceGroups.group_name == AccessRoleInstance.instance_group_name,
        AccessRoleInstance.role_group_name == RoleGroups.group_name,
        RoleGroups.role_name == unicode(role_name)).group_by(
            InstanceGroups.group_name).order_by(
                InstanceGroups.group_name).all()


def get_roles_by_group(session, group_name):
    return session.query(Roles).filter(
        RoleGroups.group_name == unicode(group_name),
        Roles.role_name == RoleGroups.role_name).order_by(
            Roles.role_name).all()


"""
Instances
"""


def add_instance(session,
                 new_agent_address,
                 new_agent_port,
                 hostname,
                 agent_key=None,
                 cpu=None,
                 memory_size=None,
                 pg_port=None,
                 pg_version=None,
                 pg_version_summary=None,
                 pg_data=None,
                 notify=False,
                 comment=None):
    try:
        instance = Instances(
            agent_address=unicode(new_agent_address),
            agent_port=int(new_agent_port),
            hostname=unicode(hostname))
        if agent_key is not None:
            instance.agent_key = unicode(agent_key)
        if cpu is not None and cpu != u'':
            instance.cpu = int(cpu)
        if memory_size is not None and memory_size != u'':
            instance.memory_size = int(memory_size)
        if pg_port is not None and pg_port != u'':
            instance.pg_port = int(pg_port)
        if pg_version is not None:
            instance.pg_version = unicode(pg_version)
        if pg_version_summary is not None:
            instance.pg_version_summary = unicode(pg_version_summary)
        if pg_data is not None:
            instance.pg_data = unicode(pg_data)
        instance.notify = bool(notify)
        instance.comment = comment
        session.add(instance)
        session.flush()
        return instance
    except IntegrityError as e:
        if e.message.find('instances_pkey') > 0:
            raise TemboardUIError(400,
                                  "Instance entry ('%s:%s') already exists." %
                                  (new_agent_address, new_agent_port))
        else:
            raise TemboardUIError(400, e.message)
    except Exception as e:
        raise TemboardUIError(400, e.message)


def get_instance(session, agent_address, agent_port):
    try:
        return session.query(Instances).options(
            joinedload(Instances.groups),
            joinedload(Instances.plugins)).filter_by(
                agent_address=unicode(agent_address),
                agent_port=agent_port).first()
    except AttributeError as e:
        raise TemboardUIError(400, "Instance entry '%s:%s' not found." %
                              (agent_address, agent_port))
    except Exception as e:
        raise TemboardUIError(e.message)


def update_instance(session,
                    agent_address,
                    agent_port,
                    new_agent_address=None,
                    new_agent_port=None,
                    agent_key=None,
                    hostname=None,
                    cpu=None,
                    memory_size=None,
                    pg_port=None,
                    pg_version=None,
                    pg_version_summary=None,
                    pg_data=None,
                    notify=True,
                    comment=None):
    try:
        instance = session.query(Instances) \
            .filter_by(
                agent_address=unicode(agent_address),
                agent_port=agent_port) \
            .first()
        if new_agent_address is not None:
            instance.agent_address = unicode(new_agent_address)
        if new_agent_port is not None:
            instance.agent_port = int(new_agent_port)
        if cpu is not None and cpu != u'':
            instance.cpu = int(cpu)
        else:
            instance.cpu = None
        if memory_size is not None and memory_size != u'':
            instance.memory_size = int(memory_size)
        else:
            instance.memory_size = None
        if pg_port is not None and pg_port != u'':
            instance.pg_port = int(pg_port)
        else:
            instance.pg_port = None
        instance.notify = bool(notify)
        instance.comment = comment

        for prop in ['agent_key', 'hostname', 'pg_version',
                     'pg_version_summary', 'pg_data']:
            if locals().get(prop) is not None:
                setattr(instance, prop, unicode(locals().get(prop)))
        session.merge(instance)
        session.flush()
        return instance
    except IntegrityError as e:
        if e.message.find('instances_pkey') > 0:
            raise TemboardUIError(400,
                                  "Instance entry ('%s:%s') already exists." %
                                  (agent_address, agent_port))
        else:
            raise TemboardUIError(400, e.message)
    except AttributeError as e:
        raise TemboardUIError(400, "Instance entry ('%s:%s') not found." %
                              (agent_address, agent_port))
    except Exception as e:
        raise TemboardUIError(400, e.message)


def delete_instance(session, agent_address, agent_port):
    from temboardui.plugins.monitoring.model.orm import (
        Host as MonitoringHost,
        Instance as MonitoringInstance,
    )

    try:
        instance = session.query(Instances).filter(
            Instances.agent_address == unicode(agent_address),
            Instances.agent_port == agent_port).one()
        session.delete(instance)
    except NoResultFound as e:
        raise TemboardUIError(400, "Instance entry ('%s:%s') not found." %
                              (agent_address, agent_port))
    except Exception as e:
        raise TemboardUIError(400, e.message)

    # Also delete any monitoring data
    # First all instance data
    try:
        monitoring_instance = session.query(MonitoringInstance) \
            .join(MonitoringHost) \
            .filter(
                MonitoringHost.hostname == instance.hostname,
                MonitoringInstance.port == instance.pg_port).one()
        session.delete(monitoring_instance)
    except NoResultFound as e:
        pass
    except Exception as e:
        raise TemboardUIError(400, e.message)

    # Then delete host data if there's no instance left referenced for this
    # host
    count = session.query(MonitoringInstance.instance_id) \
        .join(MonitoringHost) \
        .filter(MonitoringHost.hostname == instance.hostname) \
        .count()
    if count == 0:
        # Using bulk delete query here to prevent errors on not null constraint
        # on checks::host_id column (ON CASCADE DELETE not working)
        # when using session.delete(host)
        try:
            session.query(MonitoringHost) \
                .filter(MonitoringHost.hostname == instance.hostname) \
                .delete()
        except NoResultFound as e:
            pass
        except Exception as e:
            raise TemboardUIError(400, e.message)


def add_instance_in_group(session, agent_address, agent_port, group_name):
    try:
        # Create instance group if not exists
        group = Groups(group_name=unicode(group_name), group_kind=u'instance')
        session.merge(group)
        session.flush()

        instance_group = InstanceGroups(
            agent_address=unicode(agent_address),
            agent_port=agent_port,
            group_name=unicode(group_name))
        session.add(instance_group)
        session.flush()
    except IntegrityError as e:
        if e.message.find('instance_groups_group_name_fkey') > 0:
            raise TemboardUIError(
                400, "Group '%s' ('instance') does not exist." % (group_name))
        elif e.message.find('instance_groups_agent_address_fkey') > 0:
            raise TemboardUIError(400,
                                  "Instance entry ('%s:%s') does not exist." %
                                  (agent_address, agent_port))
        elif e.message.find('instance_groups_pkey') > 0:
            raise TemboardUIError(
                400, "Instance entry ('%s:%s)' already in group '%s'." %
                (agent_address, agent_port, group_name))
        else:
            raise TemboardUIError(400, e.message)
    except Exception as e:
        raise TemboardUIError(400, e.message)


def purge_instance_plugins(session, agent_address, agent_port):
    try:
        plugins = session.query(Plugins).filter(
            Plugins.agent_address == unicode(agent_address),
            Plugins.agent_port == agent_port).all()
        for plugin in plugins:
            session.delete(plugin)
    except Exception as e:
        raise TemboardUIError(400, e.message)


def add_instance_plugin(session, agent_address, agent_port, plugin_name):
    try:
        plugin = Plugins(
            agent_address=unicode(agent_address),
            agent_port=agent_port,
            plugin_name=unicode(plugin_name))
        session.add(plugin)
        session.flush()
    except IntegrityError as e:
        if e.message.find('plugins_pkey') > 0:
            raise TemboardUIError(
                400, "Plugin '%s' was already activated for this instance." %
                (plugin_name))
        elif e.message.find('plugins_agent_address_fkey') > 0:
            raise TemboardUIError(400, "Instance '%s:%s' does not exist." %
                                  (agent_address, agent_port))
        else:
            raise TemboardUIError(400, e.message)
    except Exception as e:
        raise TemboardUIError(400, e.message)


def get_instance_list(
        session, ):
    return session.query(Instances).options(
        joinedload(Instances.groups),
        joinedload(Instances.plugins)).order_by(Instances.hostname).all()


def delete_instance_from_group(session, agent_address, agent_port, group_name):
    try:
        instance_group = session.query(InstanceGroups).filter(
            InstanceGroups.agent_address == unicode(agent_address),
            InstanceGroups.agent_port == agent_port,
            InstanceGroups.group_name == unicode(group_name)).one()
        session.delete(instance_group)
    except NoResultFound as e:
        raise TemboardUIError(
            400, "Instance entry ('%s:%s)' not found in group '%s'." %
            (agent_address, agent_port, group_name))
    except Exception as e:
        raise TemboardUIError(400, e.message)


def get_instances_by_group(session, group_name):
    return session.query(Instances).options(
        joinedload(Instances.groups), joinedload(Instances.plugins)).filter(
            InstanceGroups.group_name == unicode(group_name),
            Instances.agent_address == InstanceGroups.agent_address,
            Instances.agent_port == InstanceGroups.agent_port).order_by(
                Instances.agent_address).all()


def get_groups_by_instance(session, agent_address, agent_port):
    return session.query(InstanceGroups).filter(
        InstanceGroups.agent_address == unicode(agent_address),
        InstanceGroups.agent_port == agent_port).order_by(
            InstanceGroups.group_name).all()


def get_roles_by_instance(session, agent_address, agent_port):
    return session.query(Roles) \
        .filter(
            AccessRoleInstance.role_group_name == RoleGroups.group_name,
            AccessRoleInstance.instance_group_name ==
            InstanceGroups.group_name,
            Instances.agent_address == agent_address,
            Instances.agent_port == agent_port,
            RoleGroups.role_name == Roles.role_name,
            InstanceGroups.agent_address == agent_address,
            InstanceGroups.agent_port == agent_port,
        )


def add_role_group_in_instance_group(session, role_group_name,
                                     instance_group_name):
    try:
        ari = AccessRoleInstance(
            role_group_name=unicode(role_group_name),
            instance_group_name=unicode(instance_group_name))
        session.add(ari)
        session.flush()
    except IntegrityError as e:
        if e.message.find('access_role_instance_pkey') > 0:
            raise TemboardUIError(
                400, "Group '%s' ('role') has already access to '%s'." %
                (role_group_name, instance_group_name))
        elif e.message.find(
                'access_role_instance_instance_group_name_fkey') > 0:
            raise TemboardUIError(400, "Instance group '%s' does not exist." %
                                  (instance_group_name))
        elif e.message.find('access_role_instance_role_group_name_fkey') > 0:
            raise TemboardUIError(400, "Role group '%s' does not exist." %
                                  (role_group_name))
        else:
            raise TemboardUIError(400, e.message)
    except Exception as e:
        raise TemboardUIError(400, e.message)


def delete_role_group_from_instance_group(session, role_group_name,
                                          instance_group_name):
    try:
        ari = session.query(AccessRoleInstance).filter(
            AccessRoleInstance.role_group_name == unicode(role_group_name),
            AccessRoleInstance.instance_group_name == unicode(
                instance_group_name)).one()
        session.delete(ari)
    except NoResultFound as e:
        raise TemboardUIError(
            400, "Role group '%s' not found in instance group '%s'." %
            (role_group_name, instance_group_name))
    except Exception as e:
        raise TemboardUIError(400, e.message)


def get_instances_by_role_name(session, role_name):
    return session.query(Instances).filter(
            Instances.agent_address == InstanceGroups.agent_address,
            Instances.agent_port == InstanceGroups.agent_port,
            InstanceGroups.group_name ==
            AccessRoleInstance.instance_group_name,
            AccessRoleInstance.role_group_name == RoleGroups.group_name,
            RoleGroups.role_name == unicode(role_name)).order_by(
                InstanceGroups.group_name, Instances.agent_address)


def role_name_can_access_instance(session, role_name, agent_address,
                                  agent_port):
    try:
        session.query(AccessRoleInstance).filter(
            AccessRoleInstance.instance_group_name ==
            InstanceGroups.group_name,
            AccessRoleInstance.role_group_name == RoleGroups.group_name,
            RoleGroups.role_name == unicode(role_name),
            InstanceGroups.agent_address == unicode(agent_address),
            InstanceGroups.agent_port == agent_port).one()
    except (NoResultFound, Exception):
        raise TemboardUIError(400, "You don't have access to this instance.")


def get_role_by_auth(session, role_name, role_password):
    try:
        role = session.query(Roles).filter(
            Roles.role_name == unicode(role_name),
            Roles.is_active.is_(True)).one()
        if role.role_password != unicode(role_password):
            raise TemboardUIError(400, "Wrong user/password: %s/%s" %
                                  (role_name, role_password))
        return role
    except NoResultFound as e:
        raise TemboardUIError(400, "Wrong user/password: %s/%s" %
                              (role_name, role_password))
    except Exception as e:
        raise TemboardUIError(400, e.message)


def hash_password(username, password):
    """
    Hash a password with the following formula:
        sha512(password + sha512(username))
    """
    bytes_password = hexlify(password.encode('utf-8'))
    bytes_sha_username = sha512(username.encode('utf-8')).digest()
    hpasswd = sha512(bytes_password + bytes_sha_username).digest()
    return base64.b64encode(hpasswd)


def gen_cookie(username, hash_password):
    """
    Build secure cookie content as a string containing:
        - content length (excluding length itself)
        - role_name
        - 16 first chars of the hash password

    Part of hash password is there for 2 main reasons:

        1/ If the cookie secret key is stolen, the attacker will be able to
        build a valid secure cookie, so in this situation having a
        non-predictibable variable - part of hash password -  wich can be
        checked on server side will render the attack effective less, unless
        if the attacker knows role's password.

        2/ We don't want to see the full hash password stored in a cookie.
    """
    content = "%s:%s" % (username, hash_password[:15])
    return "%s:%s" % (len(content) + 1, content)


def get_role_by_cookie(session, content):
    p_cookie_content = r'^([0-9]+):([a-z0-9_\-.]{3,16}):([A-Za-z0-9+/]{15})$'
    r_cookie_content = re.compile(p_cookie_content)
    match = r_cookie_content.match(content)
    if match:
        c_length = match.group(1)
        c_role_name = match.group(2)
        c_fp_hash_password = match.group(3)
        if int(c_length) != (len(c_role_name) + len(c_fp_hash_password) + 2):
            raise Exception("Cookie's content is corrupted.")
        try:
            role = session.query(Roles).filter(
                Roles.role_name == unicode(c_role_name),
                Roles.is_active.is_(True)).one()
        except (NoResultFound, Exception):
            raise Exception("Role '%s' not found or not active." %
                            (c_role_name))
        if role.role_password[:15] != unicode(c_fp_hash_password):
            raise Exception("Password sign not correct.")
        return role
    else:
        raise Exception("Cookie content is not valid.")


def check_role_name(role_name):
    p_role_name = r'^([a-z0-9_\-.]{3,16})$'
    r_role_name = re.compile(p_role_name)
    if not r_role_name.match(role_name):
        raise TemboardUIError(
            400, "Invalid username, must satisfy this regexp pattern: %s" %
            (p_role_name))


def check_role_email(role_email):
    p_role_email = r'^([a-zA-Z0-9_.+-]+@[a-zA-Z0-9-]+\.[a-zA-Z0-9-.]+)$'
    r_role_email = re.compile(p_role_email)
    if not r_role_email.match(role_email):
        raise TemboardUIError(400, "Invalid email address.")


def check_role_password(role_password):
    p_role_password = r'^(.{8,})$'
    r_role_password = re.compile(p_role_password)
    if not r_role_password.match(role_password):
        raise TemboardUIError(
            400, "Invalid password, it must contain at least 8 char.")


def check_role_phone(role_phone):
    p_role_phone = r'^[+][0-9]+$'
    r_role_phone = re.compile(p_role_phone)
    if not r_role_phone.match(role_phone):
        raise TemboardUIError(400, "Phone must look like +14155552671")


def check_group_name(group_name):
    p_group_name = r'^([a-z0-9_\-.]{3,16})$'
    r_group_name = re.compile(p_group_name)
    if not r_group_name.match(group_name):
        raise TemboardUIError(
            400, "Invalid group name, must satisfy this regexp pattern: %s" %
            (p_group_name))


def check_group_description(group_description):
    if len(group_description) > 255:
        raise TemboardUIError(
            400,
            "Invalid group description, must be a 256 char (max) length "
            "string."
        )


def check_agent_address(value):
    p_check = r'^([0-9a-zA-Z\-\._:]+)$'
    r_check = re.compile(p_check)
    if not r_check.match(value):
        raise TemboardUIError(400, "Invalid agent address.")


def check_agent_port(value):
    p_check = r'^([0-9]{1,5})$'
    r_check = re.compile(p_check)
    if not r_check.match(value):
        raise TemboardUIError(400, "Invalid agent port.")


def send_mail(host, port, subject, content, emails, tls=False, login=None,
              password=None, from_addr=None):

    msg = MIMEText(content, 'plain', 'utf-8')
    msg['Subject'] = subject

    try:
        if tls:
            smtp = SMTP_SSL(host, port)
        else:
            smtp = SMTP(host, port)

        if login and password:
            logger.debug("Authenticating to SMTP server as %s.", login)
            smtp.login(login, password)

        smtp.sendmail(from_addr, emails, msg.as_string())
        smtp.quit()
    except Exception as e:
        raise TemboardUIError(
            500,
            "Could not send mail; %s\n"
            "SMTP connection may be misconfigured." % e)


def send_sms(config, content, phones):
    sid = config.twilio_account_sid
    token = config.twilio_auth_token
    from_ = config.twilio_from
    uri = 'https://api.twilio.com/2010-04-01/Accounts/%s/Messages.json' % sid
    s = base64.b64encode('%s:%s' % (sid, token))

    errors = []
    for recipient in phones:
        req = urllib2.Request(url=uri)
        req.add_header('Authorization', 'Basic %s' % s)
        data = {'From': from_, 'Body': content, 'To': recipient}
        req.add_data(urllib.urlencode(data))
        try:
            urllib2.urlopen(req)
        except urllib2.HTTPError as e:
            response = json.loads(e.read())
            logger.error("Could not send SMS; %s" % response.get('message'))
            errors.append(recipient)
        except Exception as e:
            logger.error("Could not send SMS; %s" % e)
            errors.append(recipient)

    if errors:
        raise TemboardUIError(
            500,
            "Could not send SMS to %s; \n See logs for more information" %
            ', '.join(errors))
